/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.simulation.SimulationResultData;
import de.bsvrz.ars.ars.mgmt.tasks.AbstractTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveInfoQueryTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveQueryTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveRequestedTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveSettingsTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveTask;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.DeleteSimVarTask;
import de.bsvrz.ars.ars.mgmt.tasks.FlowControlTask;
import de.bsvrz.ars.ars.mgmt.tasks.HeaderRestoreTask;
import de.bsvrz.ars.ars.mgmt.tasks.LZVTask;
import de.bsvrz.ars.ars.mgmt.tasks.MirrorTask;
import de.bsvrz.ars.ars.mgmt.tasks.MultiTaskManager;
import de.bsvrz.ars.ars.mgmt.tasks.QueryInfoTask;
import de.bsvrz.ars.ars.mgmt.tasks.RequestGapTask;
import de.bsvrz.ars.ars.mgmt.tasks.RestoreTask;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.util.cron.CronDefinition;
import de.bsvrz.dav.daf.util.cron.CronScheduler;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.util.JavaSpy;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.LockSupport;

/**
 * Warteschlangen-Manager. Verteilt reinkommende Datentelegramm mit Datensätzen/Archivanfragen/Archiv-Informationsanfragen/
 * Backup-/Restore-/HeaderRestore-/LZV/Delete-/RequestGap-Aufträgen auf die jeweilige Warteschlange.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class InQueuesMgr {

	private static final int ID_ARCH_QUEUES_ONLINE = 0;
	private static final int ID_ARCH_QUEUES_REQ = 1;
	private static final int ID_NONARCH_QUEUES = 2;

	// Task Manager
	private static final int TM_QRYA_IDX = 0; /* Archivanfragen Prio A */

	private static final int TM_QRYB_IDX = 1; /* Archivanfragen Prio B */

	private static final int TM_QRYC_IDX = 2; /* Archivanfragen Prio C */

	private static final int TM_ARCINF_IDX = 3; /* Archiv-Informationsanfragen */

	// Anzahl der Query-Tasks pro Taskmanager:

	private static final int TM_QRY_A_NUM_DEFAULT = 3;  // Defaultwert gemäß Datenkatalog

	private static final int TM_QRY_B_NUM_DEFAULT = 6;  // Defaultwert gemäß Datenkatalog

	private static final int TM_QRY_C_NUM_DEFAULT = 9;  // Defaultwert gemäß Datenkatalog

	private static final int TM_ARC_INF_NUM = 4;

	private static final int NUM_OF_MULTITASK_MGR = 4;

	// Single Tasks
	public static final int ST_ARC_SETT_IDX = 0; /* Archiv-Einstellungen */

	public static final int ST_BACKUP_IDX = 1; /* Sicherung */

	public static final int ST_RESTORE_IDX = 2; /* Wiederherstellen */

	public static final int ST_HDR_REST_IDX = 3; /* Verwaltungs-Info-Abgleich */

	public static final int ST_LZV_IDX = 4; /* Löschzeit-Änderung */

	public static final int ST_DELETE_IDX = 5; /* Löschen */

	public static final int ST_REQ_GAP_IDX = 6; /* Nachfordern */

	public static final int ST_NUM_QRY_INF_IDX = 7; /* Warteschlangenanfrage */
	
	public static final int ST_MIRROR_IDX = 8; /* Mirror */

	public static final int NUM_OF_SINGLETASKS = 9;

	/** Indexe der Eingangswarteschlangen */
	public static final int Q_QRYA_IDX = 0;

	public static final int Q_QRYB_IDX = 1;

	public static final int Q_QRYC_IDX = 2;

	public static final int Q_ARCINF_IDX = 3;

	public static final int Q_BACKUP_IDX = 4;

	public static final int Q_RESTORE_IDX = 5;

	public static final int Q_HDR_REST_IDX = 6;

	public static final int Q_LZV_IDX = 7;

	public static final int Q_DELETE_IDX = 8;

	public static final int Q_REQ_GAP_IDX = 9;

	public static final int Q_ARC_SETT_IDX = 10;

	public static final int Q_SIMPARAM_IDX = 11;

	public static final int Q_NUM_QRY_INF_IDX = 12;

	public static final int NUM_OF_NONARCH_QUEUES = 13;

	private static final int[] CAPA_QUEUES = new int[NUM_OF_NONARCH_QUEUES];

	static {
		CAPA_QUEUES[Q_QRYA_IDX] = 512;
		CAPA_QUEUES[Q_QRYB_IDX] = 512;
		CAPA_QUEUES[Q_QRYC_IDX] = 512;
		CAPA_QUEUES[Q_ARCINF_IDX] = 512;
		CAPA_QUEUES[Q_BACKUP_IDX] = 512;
		CAPA_QUEUES[Q_RESTORE_IDX] = 512;
		CAPA_QUEUES[Q_HDR_REST_IDX] = 512;
		CAPA_QUEUES[Q_LZV_IDX] = 512;
		CAPA_QUEUES[Q_DELETE_IDX] = 512;
		CAPA_QUEUES[Q_REQ_GAP_IDX] = 512;
		CAPA_QUEUES[Q_ARC_SETT_IDX] = 512;
		CAPA_QUEUES[Q_SIMPARAM_IDX] = 999;
		CAPA_QUEUES[Q_NUM_QRY_INF_IDX] = 512;
	}

	/** Zahl der Tasks, die online Datensätze archivieren */
	private final int _numOfArchQueuesOnline;


	/**
	 * Initiale Maximalgröße eines Ringpuffers für aktuelle Datensätze. Die MaximalGröße kann sich auf Kosten anderer Puffer ändern. Mit dynamischen
	 * PufferGrößen kann man Lastspitzen besser abfangen.
	 */
	private final int _initialCapacityOfOnlineQueues;

	/** Kapazität des Ringpuffers für nachgelieferte Datensätze */
	private final int _initialCapacityOfRequestedQueue;

	/**
	 * Gesamtgröße (Summe) der Online-Queues
	 */
	private final int _totalCapacityOfOnlineQueues;

	/**
	 * Gesamtgröße (Summe) der Nachgefordert-Queues
	 */
	private final int _totalCapacityOfRequestedQueues;


	/** Blockgröße eines Ringpuffers für aktuelle Datensätze. In diesen Bloecken werden die Puffer verGrößert oder verkleinert. */
	private final int _onlineQueueResizeBlockSize;

	/** Ab dieser Auslastung werden Non-Write-Tasks angehalten */
	private static final double NONWRITE_SUSPEND_LIMIT = 0.8;

	/** Ab dieser Auslastung werden Non-Write-Tasks wieder fortgesetzt (mit Hysterese) */
	private static final double NONWRITE_RESUME_LIMIT = 0.7;

	/** Zählt die insgesamt in den Warteschlangen enthaltenen Datensätze */
	private final AtomicInteger onlineDataInQueues = new AtomicInteger(0);

	// Alle Ringpuffer:
	private final ArchiveRingBuffer[] archiveQueues;

	private final int[] archiveQueuesLeftout;

	@SuppressWarnings({"unchecked"})
	private final BlockingQueue<Object>[] nonArchQueues = new BlockingQueue[NUM_OF_NONARCH_QUEUES];

	// Alle direkt referenzierten Tasks (ohne die über MultiTaskManager referenzierten Tasks):
	private final ArchiveTask[] archiveTasks;

	private final AbstractTask[] singleTsks = new AbstractTask[NUM_OF_SINGLETASKS];

	private final MultiTaskManager[] tskMgrs = new MultiTaskManager[NUM_OF_MULTITASK_MGR];

	private final FlowControlTask flowCtrl;

	private final ArchiveManager archMgr;

	private final DataIdentTree didTree;

	private static final Debug logger = Debug.getLogger();

	private boolean terminated;

	// hier die Empfaenger-/Senderklassen
	private final ArchiveDataReceiver archiveDataReceiver = new ArchiveDataReceiver();

	private final QueryReceiver queryReceiver = new QueryReceiver();

	private final ArchiveSettingsReceiver arsSettingsReceiver = new ArchiveSettingsReceiver();

	private final DataAckSender dataAckSender = new DataAckSender();

	private final AtomicLong receivedCounterOnline = new AtomicLong(0);

	private final AtomicLong queueCounterRequested = new AtomicLong(0);

	private final AtomicLong queueCounterOnline = new AtomicLong(0);

	private static final String MSG_PID_QUEUE_FULL = "Warteschlangen-Kapazitaet ueberschritten";

	/** Flag ob eine Überlastsituation vorliegt und deswegen alle nicht schreibenden Tasks unterbrochen werden müssen */
	private volatile boolean suspendReadingTasks;

	private final Object suspendReadingTasksLock = new Object();

	private static final int SHIFT_QUEUES_INTERVAL = 15000;

	private final LongTermTaskStatePublisher _longTermTaskStatePublisher;

	private final Semaphore subscriptionSemaphore;

	private final Set<IdDataIdentification> _pendingSubscriptions = Collections.synchronizedSet(new HashSet<>());

	private final long _maxWaitNanosPerSubscription;

	private final long _minWaitNanosPerSubscription;

	private FreeDiskSpacePublisher _freeDiskSpacePublisher;

	/** Limit für die Queue-Größe. Wenn mehr als die angegeben Zahl an Datensätzen gequeuet sind, dann werden weitere Anmeldungen ausgebremst, bis die Queue abgearbeitet ist. */
	private final int _queueLimitForSubscriptions;

	/**
	 * Wenn true wartet das Archivsystem nicht beim Beenden auf das Leeren aller Queues.
	 */
	private volatile boolean _fastExit;

	/**
	 * Anmelden mit Verzögerung (Sliding-Window-Mechanismus)
	 * @param davCon Datenverteilerverbindung
	 * @param so Objekt
	 * @param dataDescription DataDescription
	 *
	 * @return true wenn die Anmeldung durchgeführt werden konnte, sonst false (z.B. bei Anmeldung eines Empfängers auf Senke-Aspekt)
	 */
	public boolean subscribe(final ClientDavInterface davCon, final SystemObject so, final DataDescription dataDescription) throws InterruptedException {
		IdDataIdentification identification = new IdDataIdentification(so, dataDescription);
		long waitTimeNanos = _minWaitNanosPerSubscription;
		if(countOnlineDataInQueues() >_queueLimitForSubscriptions) {
			waitTimeNanos += 10_000_000; // 10 Millisekunden extra warten, wenn die Queue mehr als halbvoll ist um das System nicht zu überlasten
		}
		if(waitTimeNanos > 0) {
			LockSupport.parkNanos(waitTimeNanos);
		}
		boolean semaphoreAcquired = false;
		if(_pendingSubscriptions.add(identification)) {
			long additionalWait = _maxWaitNanosPerSubscription - waitTimeNanos;
			if(subscriptionSemaphore.tryAcquire(additionalWait, TimeUnit.NANOSECONDS)) {
				semaphoreAcquired = true;
			}
			else {
				_pendingSubscriptions.remove(identification);
			}
		}
		try {
			davCon.subscribeReceiver(archiveDataReceiver, so, dataDescription, ReceiveOptions.delayed(), ReceiverRole.receiver());
			return true;
		}
		catch(Exception e) {
			// Anmeldung fehlgeschlagen (z.B. Empfänger auf Senke)
			if(semaphoreAcquired) {
				_pendingSubscriptions.remove(identification);
				subscriptionSemaphore.release();
			}
			logger.fine("Empfangsanmeldung fuer " + so + ", " + dataDescription + " fehlgeschlagen: " + e.getMessage());
			return false;
		}
	}

	private int totalArchQueues() {
		return _numOfArchQueuesOnline + 1;
	}

	/** Zahl der Tasks, die aktuelle Datensätze archivieren (sollte Primzahl sein) */
	public final int getNumOfArchQueuesOnline() {
		return _numOfArchQueuesOnline;
	}

	/**
	 * Erzeugt den Warteschlangen-Manager.
	 * @param aMgr  Archiv-Manager
	 * @param dTree DataIdentTree
	 * @param parameters Queue-Einstellungen
	 * */
	public InQueuesMgr(ArchiveManager aMgr, DataIdentTree dTree, final QueueParameters parameters) {
		archMgr = aMgr;
		didTree = dTree;
		_totalCapacityOfOnlineQueues = parameters.getTotalCapacityOfOnlineQueues();
		_totalCapacityOfRequestedQueues = parameters.getTotalCapacityOfRequestedQueues();
		_minWaitNanosPerSubscription = parameters.getMinWaitNanosPerSubscription();
		_maxWaitNanosPerSubscription = parameters.getMaxWaitNanosPerSubscription();

		_longTermTaskStatePublisher = new ArchiveLongTermTaskStatePublisher(archMgr, "Zustandsänderung eines Hintergrundprozesses:", "Zustände der Hintergrundprozesse:");

		// Queues (mit fester Größe):
		for(int i = 0; i < NUM_OF_NONARCH_QUEUES; i++) {
			nonArchQueues[i] = new ArrayBlockingQueue<>(CAPA_QUEUES[i]);
		}

		// Queue für Simulations-Parametrierung mit ArchivConfig verbinden:
		archMgr.getArchivConfig().setSimConfigQueue(nonArchQueues[Q_SIMPARAM_IDX]);

		assignSingleTasks();

		assignMultitasks();


		flowCtrl = new FlowControlTask();

		// parametrierbare Anzahl Query-Tasks starten:
		setQueryTaskNumbers(TM_QRY_A_NUM_DEFAULT, TM_QRY_B_NUM_DEFAULT, TM_QRY_C_NUM_DEFAULT);

		// die Zahl der ArchivInfo-Tasks ist fest:
		for(int i = 0; i < TM_ARC_INF_NUM; i++) {
			tskMgrs[TM_ARCINF_IDX].addTask(new ArchiveInfoQueryTask(tskMgrs[TM_ARCINF_IDX]));
		}

		// Queues + Tasks für Archivierung der aktuellen Datensätze
		_numOfArchQueuesOnline = parameters.getNumOfArchQueuesOnline();

		archiveQueues = new ArchiveRingBuffer[totalArchQueues()];
		archiveTasks = new ArchiveTask[totalArchQueues()];

		_initialCapacityOfOnlineQueues = _totalCapacityOfOnlineQueues / _numOfArchQueuesOnline;
		_initialCapacityOfRequestedQueue = _totalCapacityOfRequestedQueues;
		_onlineQueueResizeBlockSize = Math.max(250, _initialCapacityOfOnlineQueues / 12);

		_queueLimitForSubscriptions = Math.min(_totalCapacityOfOnlineQueues / 2, 100000);

		if(_initialCapacityOfOnlineQueues < 0 || _initialCapacityOfRequestedQueue < 0) {
			throw new IllegalArgumentException("Größe der Queues ist zu klein.");
		}

		assignArchiveTasks();
		archiveQueuesLeftout = new int[_numOfArchQueuesOnline];
		subscriptionSemaphore = new Semaphore(parameters.getSubscriptionSlidingWindowSize());
	}

	/**
	 * Archiv-Tasks initialisieren
	 */
	private void assignArchiveTasks() {
		for(int i = 0; i < _numOfArchQueuesOnline; i++) {
			archiveQueues[i] = new ArchiveRingBuffer(_onlineQueueResizeBlockSize, _initialCapacityOfOnlineQueues);
			archiveTasks[i] = new ArchiveTask(archMgr, i, archiveQueues[i]);
		}
		archiveQueues[_numOfArchQueuesOnline] = new ArchiveRingBuffer(_initialCapacityOfRequestedQueue, _initialCapacityOfRequestedQueue);
		archiveTasks[_numOfArchQueuesOnline] = new ArchiveRequestedTask(archMgr, 0, archiveQueues[_numOfArchQueuesOnline]);
	}

	/**
	 * Task-Manager (Multitasks für Archivinfo und Queries, parametrierbare Anzahl)
	 */
	private void assignMultitasks() {
		// :
		tskMgrs[TM_QRYA_IDX] = new MultiTaskManager(archMgr, nonArchQueues[Q_QRYA_IDX], "QueryPrioA");
		tskMgrs[TM_QRYB_IDX] = new MultiTaskManager(archMgr, nonArchQueues[Q_QRYB_IDX], "QueryPrioB");
		tskMgrs[TM_QRYC_IDX] = new MultiTaskManager(archMgr, nonArchQueues[Q_QRYC_IDX], "QueryPrioC");
		tskMgrs[TM_ARCINF_IDX] = new MultiTaskManager(archMgr, nonArchQueues[Q_ARCINF_IDX], "ArchivInfo");
	}

	/**
	 * Single-Tasks initialisieren
	 */
	private void assignSingleTasks() {
		singleTsks[ST_ARC_SETT_IDX] = new ArchiveSettingsTask(archMgr, nonArchQueues[Q_ARC_SETT_IDX]);
		singleTsks[ST_BACKUP_IDX] = new BackupTask(archMgr, nonArchQueues[Q_BACKUP_IDX], _longTermTaskStatePublisher);
		singleTsks[ST_RESTORE_IDX] = new RestoreTask(archMgr, nonArchQueues[Q_RESTORE_IDX], _longTermTaskStatePublisher);
		singleTsks[ST_HDR_REST_IDX] = new HeaderRestoreTask(archMgr, nonArchQueues[Q_HDR_REST_IDX]);
		singleTsks[ST_LZV_IDX] = new LZVTask(archMgr, nonArchQueues[Q_LZV_IDX], _longTermTaskStatePublisher, archMgr.getNumDeleteThreads());
		singleTsks[ST_DELETE_IDX] = new DeleteSimVarTask(archMgr, nonArchQueues[Q_DELETE_IDX]);
		singleTsks[ST_REQ_GAP_IDX] = new RequestGapTask(archMgr, nonArchQueues[Q_REQ_GAP_IDX]);
		singleTsks[ST_NUM_QRY_INF_IDX] = new QueryInfoTask(archMgr, nonArchQueues[Q_NUM_QRY_INF_IDX]);
		singleTsks[ST_MIRROR_IDX] = new MirrorTask(archMgr, _longTermTaskStatePublisher);
	}

	/**
	 * Vergrößert eine Queue (auf Kosten einer anderen, falls die Gesamtgröße überschritten ist).
	 *
	 * @param queue Zu vergrößernde Queue
	 */
	private void enlargeQueue(int queue) {
		int selectedIndex = -1;
		int maxFreeSpace = 0;
		int maxCapacity = 0;
		for(int idx = 0; idx < _numOfArchQueuesOnline; idx++) {
			if(idx == queue) continue;
			int capacity = archiveQueues[idx].maxSize();
			int freeSpace = capacity - archiveQueues[idx].size();
			if((freeSpace > maxFreeSpace) || ((freeSpace == maxFreeSpace) && (capacity > maxCapacity))) {
				maxCapacity = capacity;
				maxFreeSpace = freeSpace;
				selectedIndex = idx;
			}
		}
		if(selectedIndex >= 0) {
			assert selectedIndex != queue;
			if(archiveQueues[selectedIndex].maxSize() - archiveQueues[selectedIndex].size() >= 2 * _onlineQueueResizeBlockSize) {
				archiveQueues[selectedIndex].changeMaxSize(-_onlineQueueResizeBlockSize);
				archiveQueues[queue].changeMaxSize(_onlineQueueResizeBlockSize);
			}
			else {
				logger.warning("Zum Verkleinern ausgewählte ArchivierungsQueue (index: " + selectedIndex + ") konnte nicht verkleinert werden werden.");
			}
		}
		else {
			logger.warning("Es konnte keine ArchivierungsQueue zum Verkleinern ausgewählt werden.");
		}
	}

	/**
	 * Setzt die Anzahl der Threads für Archivanfragen auf die angegebenen Werte. Falls Threads entfernt werden, werden noch behandelte Anfragen zuende
	 * bearbeitet.
	 *
	 * @param numHi  Anzahl Tasks der Prioritaet "hoch"
	 * @param numMid Anzahl Tasks der Prioritaet "mittel"
	 * @param numLo  Anzahl Tasks der Prioritaet "niedrig"
	 */
	public void setQueryTaskNumbers(int numHi, int numMid, int numLo) {
		installQueryTask(tskMgrs[TM_QRYA_IDX], numHi);
		installQueryTask(tskMgrs[TM_QRYB_IDX], numMid);
		installQueryTask(tskMgrs[TM_QRYC_IDX], numLo);
	}

	public int getHiQueryTaskNum() {
		return tskMgrs[TM_QRYA_IDX].getTaskNum();
	}

	public int getMidQueryTaskNum() {
		return tskMgrs[TM_QRYB_IDX].getTaskNum();
	}

	public int getLoQueryTaskNum() {
		return tskMgrs[TM_QRYC_IDX].getTaskNum();
	}

	private void installQueryTask(MultiTaskManager mtm, int anz) {
		if(mtm.shouldTerminate()) return;

		int anzQTasks = mtm.getTaskNum();
		if(anz > anzQTasks) {
			for(int i = 0; i < (anz - anzQTasks); i++) {
				ArchiveQueryTask newTask = new ArchiveQueryTask(archMgr, mtm, flowCtrl);
				newTask.setMaximumRequestsPerApplication(archMgr.getMaximumQueriesPerApplication());
				mtm.addTask(newTask);
			}
		}
		else {
			for(int i = 0; i < (anzQTasks - anz); i++) {
				mtm.removeTask();
			}
		}
	}

	/**
	 * Prüft ob Tasks temporaer blockiert werden sollen und blockiert ggf. den aufrufenden Task.
	 * <p>
	 * Diese Methode und {@link #suspendNonWriteTasks()} verwenden eine reduzierte Synchronisierung um die Zahl der durchlaufenen synchronized-Bloecke
	 * gering zu halten.
	 * <p>
	 *     Diese Methode darf nur ausgeführt werden, wenn der task nicht auf eine Datenidentifikation synchronisiert ist (Deadlock-Gefahr)
	 * </p>
	 *
	 * @throws InterruptedException Thread unterbrochen
	 * @param task
	 */
	public void suspendTaskIfNecessary(final AbstractTask task) throws InterruptedException {
		archMgr.getPersMgr().assertNoLocks();

		if(suspendReadingTasks)	// synchronized nur ausfuehren falls notwendig
		{
			synchronized(suspendReadingTasksLock) {
				while(suspendReadingTasks)	// hier ist der echte Test
				{
					if(terminated || task.isTerminated()){
						// Während das Archiv oder der Task sich beendet, nicht blockieren.
						return;
					}

					suspendReadingTasksLock.wait(10000);

					// Nach 1 Sekunde erneut prüfen, ob die Tasks wieder fortgesetzt werden dürfen
					// Der Aufruf in de.bsvrz.ars.ars.mgmt.InQueuesMgr.ArchiveDataReceiver.processData() reicht nicht, da Tasks auch fortgesetzt werden sollen, wenn keine Daten eintreffen.
					suspendNonWriteTasks();
				}
			}
		}
	}

	/**
	 * Setzt das Suspend-Flag oder hebt es auf abhaengig von den beiden Grenzen.
	 * <p>
	 * Diese Methode und {@link #suspendTaskIfNecessary(AbstractTask)} verwenden eine reduzierte Synchronisierung um die Zahl der durchlaufenen synchronized-Bloecke
	 * gering zu halten. Voraußetzung: diese Methode wird nur von einem Thread aufgerufen.
	 */
	private void suspendNonWriteTasks() {
		// Zugriff auf nonWriteTasksSuspended findet nur in dieser Methode statt
		if(!suspendReadingTasks && getQueueLoad() > NONWRITE_SUSPEND_LIMIT) {
			suspendReadingTasks = true; // zu beliebigen Zeiten möglich
			logger.fine("Bearbeitung von Archivanfragen wegen Ueberlast ausgesetzt");
			//Absetzen der Betriebsmeldung
			MessageSender.getInstance().sendMessage(
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					MessageGrade.INFORMATION,
					"Die Warteschlangen sind zu " + (NONWRITE_SUSPEND_LIMIT * 100) + "% gefüllt - das Schreiben von Daten erhält eine höhere Priorität."
			);
		}
		else if(suspendReadingTasks && getQueueLoad() < NONWRITE_RESUME_LIMIT) {
			synchronized(suspendReadingTasksLock) {
				suspendReadingTasks = false;	// nur synchronisiert möglich
				suspendReadingTasksLock.notifyAll();
				logger.fine("Bearbeitung von Archivanfragen wieder aufgenommen");
				//Absetzen der Betriebsmeldung
				MessageSender.getInstance().sendMessage(
						de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
						MessageGrade.INFORMATION,
						"Die Warteschlangen sind nur noch zu " + (NONWRITE_RESUME_LIMIT * 100) + "% gefüllt. Alle Tasks arbeiten wieder mit gleicher Priorität."
				);
			}
		}
	}

	/**
	 * Hält alle lesenen Tasks direkt an, ohne weitere Bedingungen zu prüfen.
	 */
	public void suspendNonWriteTasksDirect() {
		suspendReadingTasks = true;
	}


	/**
	 * In dieser Methode wird allen Tasks (falls notwendig) die Möglichkeit gegeben, benoetigte Objekte vom DAV zu laden.
	 */
	public void getObjectsFromDav() {
		_freeDiskSpacePublisher = new FreeDiskSpacePublisher(archMgr);
		for(AbstractTask singleTsk : singleTsks) {
			singleTsk.getObjectsFromDav();
		}
		for(ArchiveTask archiveTask : archiveTasks) {
			archiveTask.getObjectsFromDav();
		}
		for(MultiTaskManager tskMgr : tskMgrs) {
			tskMgr.getObjectsFromDav();
		}
	}

	/**
	 * In dieser Methode wird allen Tasks (falls notwendig) die Möglichkeit gegeben, sich auf Objekte anzumelden. Abmelden geschieht zentral über den {@link
	 * de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager}.
	 */
	public void subscribeSettings() {
		ArchiveSettingsTask.subscribeObjects(archMgr, arsSettingsReceiver);

	}

	/**
	 * In dieser Methode wird allen Tasks (falls notwendig) die Möglichkeit gegeben, sich auf Objekte anzumelden. Abmelden geschieht zentral über den {@link
	 * de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager}.
	 */
	public void subscribeQueries() {
		ArchiveQueryTask.subscribeObjects(archMgr, queryReceiver);
	}

	/** Startet alle Tasks (Multi- und Single-Tasks) sowie die Archiv-Tasks. */
	public void startAllTasks() {
		_longTermTaskStatePublisher.setDavConnection(archMgr.getDavCon());
		resetDSCounter();
		for(int i = 0; i < NUM_OF_MULTITASK_MGR; i++) {
			tskMgrs[i].start();
		}
		for(int i = 0; i < NUM_OF_SINGLETASKS; i++) {
			singleTsks[i].start();
		}
		for(int i = 0; i < totalArchQueues(); i++) {
			archiveTasks[i].resetDSCounter();
			archiveTasks[i].start();
		}
	}

	/**
	 * Beendet alle Tasks (Multi- und Single-Tasks) sowie die Archiv-Tasks.
	 *
	 * @throws InterruptedException Thread unterbrochen
	 */
	public void stopAllTasks() throws InterruptedException {
		terminated = true;
		int totalTasks = totalArchQueues() + NUM_OF_MULTITASK_MGR + NUM_OF_SINGLETASKS;

		for(final AbstractTask singleTsk : singleTsks) {
			singleTsk.terminateTask();
		}
		for(final MultiTaskManager tskMgr : tskMgrs) {
			tskMgr.terminateTask();
		}

		waitForEmptyArchiveQueues();

		for(final ArchiveTask archiveTask : archiveTasks) {
			archiveTask.terminateTask();
		}

		// Keine weiteren Ausgaben
		archiveDataReceiver._cronScheduler.shutdownNow();

		Thread.sleep(100);
		int debugCounter = 0;
		while(true) {
			int numTasksAlive = 0;

			for(final AbstractTask singleTsk : singleTsks) {
				if(singleTsk.isAlive()) {
					numTasksAlive++;
				}
			}
			for(final MultiTaskManager tskMgr : tskMgrs) {
				if(tskMgr.isAlive()) {
					numTasksAlive++;
				}
			}
			for(final ArchiveTask archiveTask : archiveTasks) {
				if(archiveTask.isAlive()) {
					numTasksAlive++;
				}
			}

			if(numTasksAlive == 0) break;
			if(debugCounter++ == 100) {
				debugCounter = 0;
				final StringBuilder builder = new StringBuilder();
				builder.append("Tasks werden beendet...\n");
				builder.append("Es müssen noch ");
				builder.append(numTasksAlive);
				builder.append(" von ");
				builder.append(totalTasks);
				builder.append(" Tasks beendet werden.\n");
				builder.append("Verbleibende Tasks:\n");
				for(final AbstractTask singleTsk : singleTsks) {
					if(singleTsk.isAlive()) {
						builder.append(" - ").append(singleTsk.getName()).append("\n");
					}
				}
				for(final MultiTaskManager tskMgr : tskMgrs) {
					if(tskMgr.isAlive()) {
						builder.append(" - ").append(tskMgr.getName()).append("\n");
					}
				}
				for(final ArchiveTask archiveTask : archiveTasks) {
					if(archiveTask.isAlive()) {
						builder.append(" - ").append(archiveTask.getName()).append("\n");
					}
				}
				logger.warning(builder.toString());
			}
			Thread.sleep(100);
		}

		logger.info("Alle Tasks korrekt beendet");
	}

	private void waitForEmptyArchiveQueues() {
		int numOfWaitingData = 0;
		for(int i = 0; i < totalArchQueues(); i++) {
			numOfWaitingData += archiveQueues[i].size();
		}

		if(numOfWaitingData > 0) {
			logger.info(numOfWaitingData + " Datensaetze in der Warteschlange noch zu archivieren...");
			try {
				while(!_fastExit) {
					numOfWaitingData = 0;
					for(int i = 0; i < totalArchQueues(); i++) {
						numOfWaitingData += archiveQueues[i].size();
					}
					if(numOfWaitingData == 0) break;
					Thread.sleep(50);
				}
			}
			catch(InterruptedException e) {
				logger.error("Warten auf leere Archiv-Warteschlange unterbrochen.");
				return;
			}
			if(_fastExit) {
				for(ArchiveRingBuffer buffer : archiveQueues) {
					// Puffer leeren

					//noinspection SynchronizationOnLocalVariableOrMethodParameter
					synchronized(buffer) {
						//noinspection StatementWithEmptyBody
						while(buffer.popNoWait() != null) {
						}
					}
				}
				logger.error("Warten auf leere Archiv-Warteschlange wurde übersprungen.");
				return;
			}

			logger.info("Alle Datensaetze archiviert.");
		}
	}

	private static int calcArchivTaskIndex(long objID, long atgID, long aspID, int simVar, int numOfQueues) {
		return ((int)(objID + (atgID * 7) + (aspID * 17) + (simVar * 5)) & 0x7fffffff) % numOfQueues;
	}

	/**
	 * Verteilung der Datenidentifikationen auf die Online-Archiv-Tasks erfolgt anhand der IDs. Jede Datenidentifikation wird somit immer in dieselbe ArchiveQueue
	 * eingefuegt. Auf diese Weise kann die Reihenfolge des Eintreffens wesentlich einfacher erhalten werden.
	 *
	 * @param objID  Objekt-ID
	 * @param atgID  Attributgruppen-ID
	 * @param aspID  Aspekt-ID
	 * @param simVar Simulationsvariante
	 *
	 * @return Summe der IDs und simVar modulo Anzahl der ArchivTasks
	 */
	public int calcOnlineArchivTaskIndex(long objID, long atgID, long aspID, int simVar) {
		return calcArchivTaskIndex(objID, atgID, aspID, simVar, _numOfArchQueuesOnline);
	}

	/**
	 * Verteilung der Datenidentifikationen auf die Archiv-Tasks erfolgt anhand der IDs. Jede Datenidentifikation wird somit immer in dieselbe ArchiveQueue
	 * eingefuegt. Auf diese Weise kann die Reihenfolge des Eintreffens wesentlich einfacher erhalten werden.
	 *
	 * @param rd Ergebnisdatensatz
	 *
	 * @return Summe der IDs modulo Anzahl der ArchivTasks
	 */
	private int calcOnlineArchivTaskIndex(ResultData rd) {
		DataDescription dd = rd.getDataDescription();
		return calcOnlineArchivTaskIndex(
				rd.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant()
		);
	}

	/**
	 * Fuegt einen nachgeforderten Datensatz in die entsprechende Queue ein.
	 *
	 * @param archiveTime Archivzeit
	 * @param ad    Datensatz
	 */
	public void insertInArchiveQueueReq(long archiveTime, Object ad) {
		try {
			while(true) {
				if(archiveQueues[_numOfArchQueuesOnline].push(ad, archiveTime, 5000)) break;
			}
		}
		catch(InterruptedException ignored) {
		}

		if(ad instanceof ArchiveData) {
			// Nur Archivdaten zählen, keine Semaphoren (siehe RequestGapTask)
			queueCounterRequested.incrementAndGet();
		}
	}

	/**
	 * Fügt dem Simulationsvarianten-Lösch-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return {@code true} falls der Auftrag eingefügt werden konnte. {@code false} sonst.
	 */
	public boolean insertInSimVarDeleteQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_DELETE_IDX].offer(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_DELETE_IDX].size(), "Loeschen SimVar", resultData);
		return result;
	}

	/**
	 * Fügt dem Parametrierungs-Task einen weiteren Auftrag hinzu.
	 *
	 * @param resultData Auftrag. Enthält Informationen, auf welche Datenidentifikationen sich das Archivsystem anmelden soll.
	 *
	 * @return	{@code true} falls der Auftrag eingefügt werden konnte. {@code false} sonst.
	 */
	public boolean insertSimVarParam(SimulationResultData resultData) {
		boolean result = nonArchQueues[Q_SIMPARAM_IDX].offer(resultData);
		if(!result) {
			logCapaExceeded(nonArchQueues[Q_SIMPARAM_IDX].size(), "SimVar Parametrierung ändern", resultData);
		}
		return result;
	}

	/**
	 * Fügt dem Lösch-Task (automatisches Löschen) einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return {@code true} falls der Auftrag eingefügt werden konnte. {@code false} sonst.
	 */
	public boolean insertInDeleteRegularQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_LZV_IDX].offer(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_LZV_IDX].size(), "Loeschen", resultData);
		return result;
	}

	/**
	 * Fügt dem Nachfordern-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return {@code true} falls der Auftrag eingefügt werden konnte. {@code false} sonst.
	 */
	public boolean insertInRequestQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_REQ_GAP_IDX].offer(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_REQ_GAP_IDX].size(), "Nachfordern", resultData);
		return result;
	}

	/**
	 * Fügt dem LZV-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return {@code true} falls der Auftrag eingefügt werden konnte. {@code false} sonst.
	 */
	public boolean insertInLZVQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_LZV_IDX].offer(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_LZV_IDX].size(), "LZV", resultData);
		return result;
	}


	/**
	 * Fügt dem Sicherungs-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return {@code true} falls der Auftrag eingefügt werden konnte. {@code false} sonst.
	 */
	public boolean insertInBackupQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_BACKUP_IDX].offer(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_BACKUP_IDX].size(), "Sicherung", resultData);
		return result;
	}

	public void insertCloseContainer(long objID, ArchiveTask.CloseContainerObject cco) {
		int tskIdx = calcOnlineArchivTaskIndex(objID, cco.atgID, cco.aspID, 0);
		boolean pushSuccessful = false;
		long nextTimeoutTime = System.currentTimeMillis() + SHIFT_QUEUES_INTERVAL;
		try {
			while(!(pushSuccessful = archiveQueues[tskIdx].push(cco, objID, 5000))) {
				nextTimeoutTime = managePushTimeout(nextTimeoutTime, "CloseContainer-Datensatz", tskIdx);
			}
		}
		catch(InterruptedException ignored) {
		}
		accountPushResult(pushSuccessful, tskIdx, "Close-Container-Objekt", null);
	}

	private long managePushTimeout(long nextTimeoutTime, final String usageText, final int tskIdx) {
		final long now = System.currentTimeMillis();
		// Nur bis zu einer Gesamtauslastung von 83% sind bei 3000 Datensätzen pro Queue durchschnittlich noch 2 Chunks mit je 250 Datensätzen frei
		if(now >= nextTimeoutTime) {
			nextTimeoutTime = now + SHIFT_QUEUES_INTERVAL;
			final double load = getQueueLoad();
			if(tskIdx >= 0) {
				if(load <= 0.83) {
					logger.warning(
							"Ein " + usageText + " kann wegen einer vollen Archiv-Queue nicht weitergegeben werden. Die Auslastung aller "
							+ "Queues zusammen beträgt " + load * 100 + "%. Es wird versucht die Kapazität der betroffenen Queue auf Kosten "
							+ "einer anderen Queue zu vergrößern. Anschließend wird erneut versucht den betroffenen Datensatz weiterzuleiten."
					);
					enlargeQueue(tskIdx);
				}
				else {
					logger.warning(
							"Archivsystem blockiert, weil ein " + usageText + " wegen einer vollen Archiv-Queue nicht weitergegeben werden "
							+ "kann. Die Auslastung aller Queues zusammen beträgt " + load * 100 + "%."
					);
				}
			}
			else {
				logger.warning(
						"Archivsystem blockiert, weil ein " + usageText + " wegen einer vollen Archiv-Queue nicht weitergegeben werden kann."
				);
			}
		}
		return nextTimeoutTime;
	}

	private void accountPushResult(final boolean pushSuccessful, final int tskIdx, final String logTitle, final ResultData resultData) {
		if(pushSuccessful) {
			queueCounterOnline.incrementAndGet();
			onlineDataInQueues.incrementAndGet();
			if(archiveQueuesLeftout[tskIdx] > 0) {
				logCapaNoLongerExceeded(archiveQueues[tskIdx].maxSize(), logTitle, archiveQueuesLeftout[tskIdx]);
				archiveQueuesLeftout[tskIdx] = 0;
			}
		}
		else {
			// Queues voll, Datensatz verworfen, Logging starten
			if(archiveQueuesLeftout[tskIdx] == 0) {
				logCapaExceeded(archiveQueues[tskIdx].maxSize(), logTitle, resultData);
			}
			archiveQueuesLeftout[tskIdx]++;
		}
	}


	private static void logCapaExceeded(int maxsize, String qname, @Nullable ResultData rd) {
		logger.warning(
				"Kapazitaet (" + maxsize + ") der Warteschlange '" + qname + "' ueberschritten. Datensatz wurde verworfen:" + Debug.NEWLINE + (rd != null
				                                                                                                                               ? Util.rd2Str(rd)
				                                                                                                                               : "")
		);
	}

	/**
	 * Betriebsmeldung absetzen: Warteschlange voll.
	 *
	 * @param maxsize Kapazität der Warteschlange
	 * @param qname   Name der Warteschlange
	 * @param rd      Datensatz.
	 */
	protected void logSendMsgCapaExceeded(int maxsize, String qname, ResultData rd) {
		String msg =
				"Kapazitaet (" + maxsize + ") der Warteschlange '" + qname + "' ueberschritten. Datensatz wurde verworfen:" + Debug.NEWLINE + Util.rd2Str(rd);
		logger.error(msg);
		MessageSender.getInstance().sendMessage(
				MSG_PID_QUEUE_FULL, de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN, "", MessageGrade.FATAL, MessageState.MESSAGE, msg
		);
	}

	protected void logCapaNoLongerExceeded(int maxsize, String qname, int leftouts) {
		logger.info("Kapazitaet (" + maxsize + ") der Warteschlange '" + qname + "' wieder ausreichend. " + leftouts + " Datensaetze wurde verworfen.");
	}

	/** @return Status der Warteschlangen für aktuelle Datensätze */
	public float[][] getArchiveQueuesOnlineStatus() {
		return getQueueStatus(ID_ARCH_QUEUES_ONLINE);
	}

	/** @return Status der Warteschlangen für nachgelieferte Datensätze */
	public float[][] getArchiveQueuesRequestedStatus() {
		return getQueueStatus(ID_ARCH_QUEUES_REQ);
	}

	/** @return Status der Warteschlangen für eingehende, nicht zu archivierende Objekte */
	public float[][] getNonArchiveQueuesStatus() {
		return getQueueStatus(ID_NONARCH_QUEUES);
	}

	private float[][] getQueueStatus(int whichQueues) {
		float[][] erg;
		switch(whichQueues) {
			case ID_ARCH_QUEUES_ONLINE:
				erg = new float[_numOfArchQueuesOnline][2];
				break;
			case ID_ARCH_QUEUES_REQ:
				erg = new float[1][2];
				break;
			case ID_NONARCH_QUEUES:
				erg = new float[NUM_OF_NONARCH_QUEUES][2];
				break;
			default:
				return new float[][]{};
		}

		for(int i = 0; i < erg.length; i++) {
			int size;
			int maxSize;
			switch(whichQueues) {
				case ID_ARCH_QUEUES_ONLINE:
					size = archiveQueues[i].size();
					maxSize = archiveQueues[i].maxSize();
					break;
				case ID_ARCH_QUEUES_REQ:
					size = archiveQueues[_numOfArchQueuesOnline + i].size();
					maxSize = archiveQueues[_numOfArchQueuesOnline + i].maxSize();
					break;
				default:
					size = nonArchQueues[i].size();
					maxSize = CAPA_QUEUES[i];
					break;
			}
			erg[i][0] = size;
			erg[i][1] = (float)size / maxSize;
		}
		return erg;
	}


	/** @return Gesamtkapazitaet aller Ringpuffers für aktuelle Datensätze */
	public int getTotalOnlineQueuesCapa() {
		return _numOfArchQueuesOnline * _initialCapacityOfOnlineQueues;
	}

	/** @return Auslastungsgrad der Warteschlangen. Anhand dieses Wertes wird ermittelt, ob die Bearbeitung nicht schreibender Tasks ausgesetzt wird. */
	public double getQueueLoad() {
		return (double)countOnlineDataInQueues() / (double)getTotalOnlineQueuesCapa();
	}

	public int countOnlineDataInQueues() {
		// 		return Arrays.stream(archiveQueues).mapToInt(RingBuffer::size).sum();
		return onlineDataInQueues.intValue();
	}

	/** Setzt alle Zähler auf 0. */
	public void resetDSCounter() {
		receivedCounterOnline.set(0);
		queueCounterRequested.set(0);
		queueCounterOnline.set(0);
		Arrays.stream(archiveTasks).forEach(ArchiveTask::resetDSCounter);
	}

	public void decrOnlineDataInQueues() {
		onlineDataInQueues.decrementAndGet();
	}


	/**
	 * Nur für Testzwecke: Liefert den gewuenschten ArchiveTask
	 *
	 * @param objId  Objekt-ID
	 * @param atgId  Attributgruppen-ID
	 * @param aspId  Aspekt-ID
	 * @param simVar Simulationsvariante
	 *
	 * @return ArchiveTask
	 *
	 * @see #calcArchivTaskIndex(long,long,long, int, int)
	 */
	public ArchiveTask getOnlineArchiveTask(long objId, long atgId, long aspId, int simVar) {
		return archiveTasks[calcArchivTaskIndex(objId, atgId, aspId, simVar, _numOfArchQueuesOnline)];
	}

	/**
	 * Liefert den Task, der für die Bearbeitung der Archiveinstellungen verantwortlich ist. Dies ist notwendig, damit der {@link ArchiveManager} beim Start einen
	 * {@link de.bsvrz.ars.ars.mgmt.tasks.ArchiveSettingsTask.ArSSettingListener} einhängen und darauf warten kann, dass die Bearbeitung der Archiveinstellungen
	 * abgeschlossen ist. Das Archivsystem wartet beim Start auf die Archiv-Einstellungen, da es nicht sinnvoll ist, die ersten Sekunden mit Default-Werten
	 * loszulaufen.
	 *
	 * @return Den Task, der für die Bearbeitung der Archiveinstellungen verantwortlich ist.
	 */
	public ArchiveSettingsTask getArchiveSettingsTask() {
		return (ArchiveSettingsTask)getSingleTask(InQueuesMgr.ST_ARC_SETT_IDX);
	}


	/**
	 * Liefert den gewuenschten SingleTask (um etwa Listener einzuhängen)
	 *
	 * @param taskIndex Index der internen SingleTask-Liste
	 *
	 * @return SingleTask
	 */
	public AbstractTask getSingleTask(int taskIndex) {
		return singleTsks[taskIndex];
	}

	/**
	 * Die Klasse {@link ArchivConfig} meldet die zu archivierenden Daten mit dem Objekt als Empfaenger an, das von dieser Methode geliefert wird.
	 *
	 * @return Empfaengerobjekt für zu archivierende Daten.
	 */
	public ArchiveDataReceiver getArchiveDataReceiver() {
		return archiveDataReceiver;
	}

	/**
	 * Die Klasse {@link ArchivConfig} meldet die Quittungen für die zu archivierenden Daten mit dem Objekt als Sender an, das von dieser Methode geliefert wird.
	 *
	 * @return Senderobjekt für die Quittungen der zu archivierenden Daten.
	 */
	public DataAckSender getDataAckSender() {
		return dataAckSender;
	}

	/**
	 * Aktualisiert den freien Speicherplatz
	 * @param freeDiskSpace freier Speicherplatz in Bytes
	 */
	public void setFreeDiskSpace(final long freeDiskSpace) {
		if(_freeDiskSpacePublisher != null) {
			_freeDiskSpacePublisher.sendFreeSpace(freeDiskSpace);
		}
	}

	/**
	 * Initiale Maximalgröße eines Ringpuffers für aktuelle Datensätze. Die MaximalGröße kann sich auf Kosten anderer Puffer ändern. Mit dynamischen
	 * PufferGrößen kann man Lastspitzen besser abfangen.
	 */
	public int getInitialCapacityOfOnlineQueues() {
		return _initialCapacityOfOnlineQueues;
	}

	/** Kapazität des Ringpuffers für nachgelieferte Datensätze */
	public int getInitialCapacityOfRequestedQueue() {
		return _initialCapacityOfRequestedQueue;
	}

	/**
	 * Gesamtgröße (Summe) der Online-Queues
	 */
	public int getTotalCapacityOfOnlineQueues() {
		return _totalCapacityOfOnlineQueues;
	}

	/**
	 * Gesamtgröße (Summe) der Nachgefordert-Queues
	 */
	public int getTotalCapacityOfRequestedQueues() {
		return _totalCapacityOfRequestedQueues;
	}

	/**
	 * Größe der Blocks, in denen die Online-Queues vergrößert oder verkleinert werden
	 * @return
	 */
	public int getOnlineQueueResizeBlockSize() {
		return _onlineQueueResizeBlockSize;
	}

	/**
	 * Setzt, ob sich der InQueuesMgr schnell terminieren soll. Bei einer schnellen Terminierung wird nicht
	 * auf die Abarbeitugn aller Queues gewartet.
	 * @param quickExit true wenn schnell terminiert werden soll, sonst false
	 */
	public void setFastExit(final boolean quickExit) {
		_fastExit = quickExit;
	}

	/** Empfaengerklasse für Archiveinstellungen */
	public class ArchiveSettingsReceiver extends DataReceiver {

		@Override
		public void processData(ResultData rd) {
			logger.finer("Archiveinstellung erhalten: ", rd);
			if(rd.hasData()) {
				if(!nonArchQueues[Q_ARC_SETT_IDX].offer(rd)) logSendMsgCapaExceeded(nonArchQueues[Q_ARC_SETT_IDX].size(), "Archiveinstellungen", rd);
			}
		}
	}

	/** Empfaengerklasse für Archivanfragen */
	public class QueryReceiver extends DataReceiver {

		private Deserializer deserializer;

		@Override
		public void processData(ResultData rd) {
			if(rd.hasData()) {
				int msgType = rd.getData().getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).intValue();
				switch(msgType) {
					case MessageType.QUERY:  // Anfrage
						ArchiveQueryPriority prio;
						try {
							prio = getQueryPrio(rd);
						}
						catch(Exception e) {
							logger.warning("Die Prioritaet einer Archivanfrage konnte nicht ermittelt werden: " + rd, e.getMessage());
							prio = ArchiveQueryPriority.MEDIUM;
						}
						int queueIdx = ArchiveQueryPriority.HIGH == prio ? Q_QRYA_IDX : ArchiveQueryPriority.MEDIUM == prio ? Q_QRYB_IDX : Q_QRYC_IDX;
						if(!nonArchQueues[queueIdx].offer(rd)) logSendMsgCapaExceeded(nonArchQueues[queueIdx].size(), "Archivanfrage " + prio, rd);
						break;
					case MessageType.QUERY_INFO:		  // Informationsanfrage:
						if(!nonArchQueues[Q_ARCINF_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_ARCINF_IDX].size(), "Archiv-Informationsanfrage ", rd);
						break;
					case MessageType.BACKUP:			  // Sicherungsauftrag:
						if(!nonArchQueues[Q_BACKUP_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_BACKUP_IDX].size(), "Sicherung", rd);
						break;
					case MessageType.DELETE:			  // Loeschauftrag für eine Simulationsvariante:
						if(!nonArchQueues[Q_DELETE_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_DELETE_IDX].size(), "Loeschen SimVar", rd);
						break;
					case MessageType.DELETION_TIME:	 // Loeschzeitpunkt-Änderung:
						if(!nonArchQueues[Q_LZV_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_LZV_IDX].size(), "Loeschzeitpunkt", rd);
						break;
					case MessageType.RESTORE:			// Wiederherstellung:
						if(!nonArchQueues[Q_RESTORE_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_RESTORE_IDX].size(), "Wiederherstellung", rd);
						break;
					case MessageType.HEADER_RESTORE:	// Verwaltungsinformationsabgleich:
						if(!nonArchQueues[Q_HDR_REST_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_HDR_REST_IDX].size(), "Verwaltungsinfoabgleich", rd);
						break;
					case MessageType.STREAM_CONTROL:	// Flusskontrolle:
						flowCtrl.processFlowCtrl(rd);
						break;
					case MessageType.REQUEST_DID:
					case MessageType.REQUEST_AUTOM:
						if(!nonArchQueues[Q_REQ_GAP_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_REQ_GAP_IDX].size(), "Nachforderungsauftrag", rd);
						break;
					case MessageType.NUM_QUERIES_INFO:
						if(!nonArchQueues[Q_NUM_QRY_INF_IDX].offer(rd)) logCapaExceeded(nonArchQueues[Q_NUM_QRY_INF_IDX].size(), "Anfrage Warteschlangenkapazität", rd);
						break;

					default:
						logger.warning("Nachrichtentyp erhalten, der nicht bearbeitet werden kann: " + msgType + " aus: " + Util.rd2Str(rd));
				}
			}
		}

		private ArchiveQueryPriority getQueryPrio(ResultData rd) throws IOException, NoSuchVersionException {
			byte[] data = rd.getData().getUnscaledArray(PidQuery.ATT_DATA_NAME).getByteArray();
			InputStream stream = new ByteArrayInputStream(data, 4, data.length - 4);
			int deserVersion = Util.getSerVersion(data);
			if(deserializer == null || deserializer.getVersion() != deserVersion) {
				deserializer = SerializingFactory.createDeserializer(deserVersion, stream);
			}
			else {
				deserializer.setInputStream(stream);
			}
			return ArchiveQueryPriority.getInstance(deserializer.readInt()); // Prioritaet
		}
	}

	/** Empfaengerklasse für Archivdaten */
	public class ArchiveDataReceiver extends DataReceiver {

		public static final String MSG_PID_ATBACKSTEP_SUSPEND = "Archivzeit-Ruecksprung (suspend)";

		public static final String MSG_PID_ATBACKSTEP_RESUME = "Archivzeit-Ruecksprung (resume)";

		public static final String MSG_PID_ATFWSTEP_SUSPEND = "Archivzeit-Vorwaertssprung (suspend)";

		public static final String MSG_PID_ATFWSTEP_RESUME = "Archivzeit-Vorwaertssprung (resume)";

		private boolean suspendArchMessageSentBS, resumeArchMessageSentBS;	// Backstep

		private boolean suspendArchMessageSentFS, resumeArchMessageSentFS;	// Forward Step

		/**
		 * Wird auf false gesetzt, nachdem der erste Datensatz empfangen wurde. Wird verwendet, damit Vorwärtssprungs auch nach einem Neustart ses Archivsystems
		 * korrekt erkannt wird
		 */
		private boolean firstData = true;

		/** max. Archivzeitsprung vorwärts (siehe TAnfArS 5.1.2.4.3.3) */
		private long tMaxATimeForwardStep = 3600 * 1000;	// Defaultwert 1 Stunde gemäß DatK

		private volatile long _statisticLastTotalReceivedDataSetCount;

		private volatile long _statisticLastTotalQueuedDataSetCount;

		private volatile long _statisticLastTotalSuccessedDataSetCount;

		private volatile long _statisticLastTotalFailedDataSetCount;

		private volatile long _statisticLastTotalQueuedDataSetCountRequested;

		private volatile long _statisticLastTotalSuccessedDataSetCountRequested;

		private volatile long _statisticLastTotalFailedDataSetCountRequested;

		private long[] _statisticLastCacheCounts = {0, 0, 0, 0};
		
		private final CronScheduler _cronScheduler;

		public ArchiveDataReceiver() {
			_cronScheduler = new CronScheduler(1, r -> {
				Thread statistics = new Thread(r, "Statistics");
				statistics.setDaemon(true);
				return statistics;
			});
			_cronScheduler.schedule(this::printStatistics, CronDefinition.EVERY_MINUTE);
		}

		@Override
		public void update(final ResultData[] rds) {
			for(ResultData rd : rds) {
				if(!_pendingSubscriptions.isEmpty() && _pendingSubscriptions.remove(new IdDataIdentification(rd.getObject(), rd.getDataDescription()))){
					subscriptionSemaphore.release();
				}
			}
			super.update(rds);
		}

		@Override
		public void processData(ResultData rd) {
			long aTime = ArchiveManager.archiveTime();

			receivedCounterOnline.incrementAndGet();

			// Zu Testzwecken
			if(ArchiveManager.ONLY_DATA && !rd.hasData()) return;

			DataIdentNode din = didTree.get(rd);
			if(din.arSParamIsArchivieren() && din.isArSParameterized()) {
				final long lastArchiveTime = ArchiveTask.getLastArchiveTime();
				// Folgende Abfrage sorgt dafür, dass kurze Rücksprünge der Zeit (bis 10 Sekunden) toleriert werden, in dem einfach die entsprechende
				// Zeit abgewartet wird und die aktuelle Zeit erneut abgefragt wird.
				if(aTime < lastArchiveTime) {
					long delta = lastArchiveTime - aTime;
					if(delta < 10000) {
						try { Thread.sleep(delta); } catch(InterruptedException ignored) { }
						aTime = ArchiveManager.archiveTime();
					}
				}
				if(!testForArchiveTimeBackstep(aTime, lastArchiveTime) && !testForArchiveTimeForwardStep(aTime, lastArchiveTime)) {
					putDataInQueue(aTime, rd);
					ArchiveTask.setLastArchiveTime(aTime);

					// Auf Überlast prüfen und ggf. alle nicht schreibenden Tasks anhalten
					suspendNonWriteTasks();
				}
			}
		}

		public void printStatistics() {
			StringBuilder message = new StringBuilder();
			message.append("Aktivität der Archiv-Queues im letzten Ausgabeintervall:");

			final long totalReceivedDataSetCount = getReceivedCountOnline();
			message.append(String.format("%n%20d Datensätze empfangen", totalReceivedDataSetCount - _statisticLastTotalReceivedDataSetCount));
			_statisticLastTotalReceivedDataSetCount = totalReceivedDataSetCount;

			final long totalQueuedDataSetCount = getQueuedCountOnline();
			message.append(String.format("%n%20d Datensätze weitergeleitet", totalQueuedDataSetCount - _statisticLastTotalQueuedDataSetCount));
			_statisticLastTotalQueuedDataSetCount = totalQueuedDataSetCount;

			final long totalSuccessedDataSetCount = getSuccessCountOnline();
			message.append(String.format("%n%20d Datensätze erfolgreich archiviert", totalSuccessedDataSetCount - _statisticLastTotalSuccessedDataSetCount));
			_statisticLastTotalSuccessedDataSetCount = totalSuccessedDataSetCount;

			final long totalFailedDataSetCount = getFailedCountOnline();
			message.append(String.format("%n%20d Datensätze nicht erfolgreich archiviert", totalFailedDataSetCount - _statisticLastTotalFailedDataSetCount));
			_statisticLastTotalFailedDataSetCount = totalFailedDataSetCount;

			final long totalQueuedDataSetCountRequested = getQueuesCountRequested();
			message.append(String.format("%n%20d nachgeforderte Datensätze empfangen", totalQueuedDataSetCountRequested - _statisticLastTotalQueuedDataSetCountRequested));
			_statisticLastTotalQueuedDataSetCountRequested = totalQueuedDataSetCountRequested;

			final long totalSuccessedDataSetCountRequested = getSuccessCountRequested();
			message.append(String.format("%n%20d nachgeforderte Datensätze erfolgreich archiviert", totalSuccessedDataSetCountRequested - _statisticLastTotalSuccessedDataSetCountRequested));
			_statisticLastTotalSuccessedDataSetCountRequested = totalSuccessedDataSetCountRequested;

			final long totalFailedDataSetCountRequested = getFailedCountRequested();
			message.append(String.format("%n%20d nachgeforderte Datensätze nicht erfolgreich archiviert", totalFailedDataSetCountRequested - _statisticLastTotalFailedDataSetCountRequested));
			_statisticLastTotalFailedDataSetCountRequested = totalFailedDataSetCountRequested;

			message.append(String.format("%n%20d Datensätze noch in den Queues befindlich", countOnlineDataInQueues()));
			message.append(String.format("%n%19d%% Auslastung der Queues", (int)(getQueueLoad() * 100.0)));

			message.append("\nAktivität der Zwischenspeicherung in geschlossenen Cache-Objekten: ");
			final long[] cacheCounts = CacheManager.getInstance().getCounts();
			message.append(String.format("%n%20d Datei-Schreibvorgänge", cacheCounts[0] - _statisticLastCacheCounts[0]));
			message.append(String.format("%n%20d verarbeitete Datenblöcke", cacheCounts[1] - _statisticLastCacheCounts[1]));
			message.append(String.format("%n%20d zwischengespeicherte Datenblöcke", cacheCounts[2] - _statisticLastCacheCounts[2]));
			message.append(String.format("%n%20d nicht zwischengespeicherte Datenblöcke", cacheCounts[3] - _statisticLastCacheCounts[3]));
			_statisticLastCacheCounts = cacheCounts;

			message.append("\nAktueller Speicherverbrauch: ");
			PersistenceManager.Statistics statistics = archMgr.getPersMgr().getStatistics();
			Runtime runtime = Runtime.getRuntime();
			long usedMemory = runtime.totalMemory() - runtime.freeMemory();
			message.append(String.format("%n%20.3f MB insgesamt, davon", usedMemory / 1048576.0));
			message.append(String.format("%n%20.3f MB zur Synchronisierung", statistics.getLockMemory() / 1048576.0));
			message.append(String.format("%n%20.3f MB für %d Datenknoten (davon %.0f%% parametriert)", statistics.getNodeMemory() / 1048576.0, statistics.getNumNodes(), 100.0 * statistics.getNumParameterizedNodes() / statistics.getNumNodes()));
			message.append(String.format("%n%20.3f MB für %d offene Container", statistics.getOpenContainerMemory() / 1048576.0, statistics.getNumOpenContainerData()));
			message.append(String.format("%n%20.3f MB zwischengespeichert für Archivierungs-Cache (davon %.0f%% ungenutzt)",
			                             statistics.getCachedMemory().getTotalBytesUsed() / 1048576.0,
			                             100.0 - 100.0 * (double)statistics.getCachedMemory().getCachedBytes() / statistics.getCachedMemory().getTotalBytesUsed())
			);

			logger.info(message.toString());
		}

		private boolean testForArchiveTimeForwardStep(long aTime, final long lastArchiveTime) {
			if(firstData) {	//nach einem Neustart des Archivsystems soll die Überprüfung erst beim zweiten Archivdatensatz vorgenommen werden
				firstData = false;
				return false;
			}
			else if(aTime - lastArchiveTime > tMaxATimeForwardStep) {
				if(!suspendArchMessageSentFS) {
					sendArchiveSuspendMsgFS(aTime);
					suspendArchMessageSentFS = true;
					resumeArchMessageSentFS = false;
				}
				return true;
			}
			else {
				if(suspendArchMessageSentFS && !resumeArchMessageSentFS) {
					sendArchiveResumeMsgFS(aTime);
					resumeArchMessageSentFS = true;
					suspendArchMessageSentFS = false;
				}
				return false;
			}
		}

		/**
		 * Prüft ob ein Rücksprung der Archivzeit aufgetreten ist und sendet entsprechende Betriebsmeldungen.
		 *
		 * @param aTime aktuelle Archivzeit
		 *
		 * @param lastArchiveTime
		 * @return Wahr, wenn ein Rücksprung aufgetreten ist, falsch sonst
		 */
		private boolean testForArchiveTimeBackstep(long aTime, final long lastArchiveTime) {
			if(aTime < lastArchiveTime) {
				if(!suspendArchMessageSentBS) {
					sendArchiveSuspendMsgBS(aTime);
					suspendArchMessageSentBS = true;
					resumeArchMessageSentBS = false;
				}
				return true;
			}
			else {
				if(suspendArchMessageSentBS && !resumeArchMessageSentBS) {
					sendArchiveResumeMsgBS(aTime);
					resumeArchMessageSentBS = true;
					suspendArchMessageSentBS = false;
				}
				return false;
			}
		}

		/**
		 * Legt das Datum in die passende Warteschlange. Falls die voll ist, wird einmal eine Log-Message ausgegeben. Wenn sie nicht mehr voll ist, wird ebenfalls
		 * einmal eine Log-Message ausgegeben.
		 *
		 * @param archiveTime Archivzeit des zu archivierenden Datums
		 * @param resultData  zu archivierendes Datum
		 */
		private void putDataInQueue(long archiveTime, ResultData resultData) {
			int tskIdx = calcOnlineArchivTaskIndex(resultData);

			boolean pushSuccessful = false;
			try {
				long nextTimeoutTime = archiveTime + SHIFT_QUEUES_INTERVAL;
				while(!(pushSuccessful = archiveQueues[tskIdx].push(resultData, archiveTime, 5000))) {
					nextTimeoutTime = managePushTimeout(nextTimeoutTime, "empfangener Online-Datensatz", tskIdx);
				}
			}
			catch(InterruptedException ignored) {
			}
			accountPushResult(pushSuccessful, tskIdx, "Archivierung", resultData);
		}

		private void sendArchiveSuspendMsgFS(long aTime) {
			String msg = "Es ist ein Vorwaertssprung der Archivzeit aufgetreten: die Archivzeit des soeben erhaltenen Datensatzes ("
			             + Util.timestrMillisFormatted(aTime) + ") liegt mehr als " + tMaxATimeForwardStep / 1000 + "sec ("
			             + Util.relTimestrMillis(tMaxATimeForwardStep) + ") nach der Archivzeit des zuletzt archivierten Datensatzes ("
			             + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime())
			             + "). Das Archivsystem stellt die Archivierung ein, bis die Archivzeit wieder innerhalb der zulaessigen Spanne liegt.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATFWSTEP_SUSPEND,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.FATAL,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		private void sendArchiveResumeMsgFS(long aTime) {
			String msg = "Der Vorwaertssprung der Archivzeit ist behoben: die Archivzeit des soeben erhaltenen Datensatzes ("
			             + Util.timestrMillisFormatted(aTime) + ") liegt wieder innerhalb der zulaessigen Spanne von "
			             + tMaxATimeForwardStep / 1000 + "sec (" + Util.relTimestrMillis(tMaxATimeForwardStep)
			             + ") seit der Archivzeit des zuletzt archivierten Datensatzes (" + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime())
			             + ". Das Archivsystem hat die Archivierung wieder aufgenommen.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATFWSTEP_RESUME,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.INFORMATION,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		private void sendArchiveSuspendMsgBS(long aTime) {
			String msg = "Es ist ein Ruecksprung der Archivzeit aufgetreten: die Archivzeit des zuletzt archivierten " + "Datensatzes ("
			             + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime()) + ") liegt nach der Archivzeit des soeben erhaltenen Datensatzes ("
			             + Util.timestrMillisFormatted(aTime) + "). Das Archivsystem stellt die Archivierung ein, bis die Archivzeit wieder nach der Archivzeit"
			             + " des zuletzt archivierten Datensatzes liegt.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATBACKSTEP_SUSPEND,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.FATAL,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		private void sendArchiveResumeMsgBS(long aTime) {
			String msg = "Der Ruecksprung der Archivzeit ist behoben: die Archivzeit des zuletzt archivierten " + "Datensatzes ("
			             + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime())
			             + ") liegt wieder vor der Archivzeit des soeben erhaltenen Datensatzes (" + Util.timestrMillisFormatted(aTime)
			             + "). Das Archivsystem hat die Archivierung wieder aufgenommen.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATBACKSTEP_RESUME,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.INFORMATION,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		/**
		 * Setzt den maximal erlaubten Vorwärtssprung der Archivzeit.
		 *
		 * @param tMaxSeconds max. erlaubter Sprung (in Sekunden)
		 */
		public void setTMaxATimeForwardStep(long tMaxSeconds) {
			tMaxATimeForwardStep = tMaxSeconds * 1000;
		}

		public long getTMaxATimeForwardStep() {
			return tMaxATimeForwardStep;
		}
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} von {@link #getArchiveDataReceiver()} empfangenen Datensätze, egal ob diese
	 *         archiviert werden oder nicht. Kann zur Test-Synchronisation verwendet werden.
	 */
	public long getReceivedCountOnline() {
		return receivedCounterOnline.longValue();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Datensätze, die in eine Queue eingefügt wurden.
	 */
	public long getQueuedCountTotal() {
		return getQueuedCountOnline() + getQueuesCountRequested();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Online-Datensätze, die in eine Queue eingefügt wurden.
	 */
	public long getQueuedCountOnline() {
		return queueCounterOnline.get();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} nachgeforderten Datensätze, die in eine Queue eingefügt wurden.
	 */
	public long getQueuesCountRequested() {
		return queueCounterRequested.get();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Datensätze, die nicht erfolgreich archiviert wurden
	 */
	public long getFailedCountTotal() {
		return Arrays.stream(archiveTasks).mapToLong(ArchiveTask::getFailedCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Datensätze, die erfolgreich archiviert wurden
	 */
	public long getSuccessCountTotal() {
		return Arrays.stream(archiveTasks).mapToLong(ArchiveTask::getSuccessCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Online-Datensätze, die nicht erfolgreich archiviert wurden
	 */
	public long getFailedCountOnline() {
		return Arrays.stream(archiveTasks).filter(it -> it.getClass() == ArchiveTask.class).mapToLong(ArchiveTask::getFailedCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Online-Datensätze, die erfolgreich archiviert wurden
	 */
	public long getSuccessCountOnline() {
		return Arrays.stream(archiveTasks).filter(it -> it.getClass() == ArchiveTask.class).mapToLong(ArchiveTask::getSuccessCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} nachgeforderten Datensätze, die nicht erfolgreich archiviert wurden
	 */
	public long getFailedCountRequested() {
		return Arrays.stream(archiveTasks).filter(it -> it.getClass() == ArchiveRequestedTask.class).mapToLong(ArchiveTask::getFailedCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} nachgeforderten Datensätze, die erfolgreich archiviert wurden
	 */
	public long getSuccessCountRequested() {
		return Arrays.stream(archiveTasks).filter(it -> it.getClass() == ArchiveRequestedTask.class).mapToLong(ArchiveTask::getSuccessCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Datensätze, die verarbeitet wurden
	 */
	public long getProcessedCountTotal() {
		return Arrays.stream(archiveTasks).mapToLong(ArchiveTask::getProcessedCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} Online-Datensätze, die verarbeitet wurden
	 */
	public long getProcessedCountOnline() {
		return Arrays.stream(archiveTasks).filter(it -> it.getClass() == ArchiveTask.class).mapToLong(ArchiveTask::getProcessedCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} nachgeforderten Datensätze, die verarbeitet wurden
	 */
	public long getProcessedCountRequested() {
		return Arrays.stream(archiveTasks).filter(it -> it.getClass() == ArchiveRequestedTask.class).mapToLong(ArchiveTask::getProcessedCount).sum();
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} erfolgreich verarbeiteten Close-Container-Datensätze. Kann zur
	 *         Test-Synchronisation verwendet werden.
	 */
	public long getCloseContainerSuccess() {
		return Arrays.stream(archiveTasks).mapToLong(ArchiveTask::getCloseContainerSuccess).sum();
	}

	/**
	 * Abstrakte Klasse, die zum Empfang bestimmter Daten (z.B. von Archiv-Queries) verwendet werden kann. Auf diese Weise muss das eingehende Datum nicht mehr
	 * anhand von IDs unterschieden und der passenden Warteschlange zugewiesen werden.
	 */
	public abstract class DataReceiver implements ClientReceiverInterface {

		private final String logPrefix = getClass().getSimpleName() + " update()";

		@Override
		public void update(ResultData[] rds) {            
			logger.finer(logPrefix, "ResultData[] " + ("mit " + rds.length + " Datensaetzen") + " erhalten");

			for(final ResultData rd : rds) {
				// *** NUR FÜR TESTZWECKE
				if(ArchiveManager.DATAINDEX_OFFSET != 0) {
					logger.warning("Datenindex-Offset aktiviert! (" + ArchiveManager.DATAINDEX_OFFSET + ")");
					try {
						JavaSpy.setFieldValue(rd, "dataIndex", rd.getDataIndex() + ArchiveManager.DATAINDEX_OFFSET);
					}
					catch(Exception e) {
						logger.error("Verstellen des Datenindex zu Testzwecken fehlgeschlagen: " + e.getMessage());
					}
				}
				// ***

				if(terminated) {
//					logger.info("Archivsystem wird gerade beendet, Daten werden nicht mehr bearbeitet.");
					return;
				}
				// TAnfArS 5.1.2.4.3.7: initiale "keine Quelle"-DS ignorieren
				if(!isInitialNoSourceDS(rd)) {
					processData(rd);
				}
			}
		}

		/**
		 * Prüft, ob es sich bei dem Datensatz um einen initialen "Keine Quelle"-Datensatz handelt, der vom DaV erzeugt wird, sobald sich ein Empfaenger angemeldet
		 * hat, ohne dass eine Quelle existiert.
		 *
		 * @param rd Datensatz
		 *
		 * @return Wahr, falls initialer "Keine Quelle"-Datensatz, falsch sonst
		 */
		private boolean isInitialNoSourceDS(ResultData rd) {
			return Util.dIdxSrcSubscrTime(rd) == 0 && Util.dIdxLfdnr(rd) == 0;
		}

		public abstract void processData(ResultData rd);
	}

	public class DataAckSender implements ClientSenderInterface {

		public void sendAck(ResultData rd, long qAspID) throws DataNotSubscribedException, SendSubscriptionNotConfirmed {
			ResultData qRd = new ResultData(
					rd.getObject(),
					new DataDescription(rd.getDataDescription().getAttributeGroup(), archMgr.getAsp(qAspID)),
					ArchiveManager.archiveTime(),
					rd.getData()
			);
			archMgr.getDavCon().sendData(new ResultData[]{qRd});
		}

		@Override
		public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
			return true;
		}

		@Override
		public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
			if(state == ClientSenderInterface.STOP_SENDING_NOT_A_VALID_SUBSCRIPTION) {
				logger.warning(
						"Sendesteuerung fuer Quittung '" + object + ", " + dataDescription + "': STOP_SENDING_NOT_A_VALID_SUBSCRIPTION" + Debug.NEWLINE
						+ "Moeglicherweise ist unter dieser Datenidentifikation bereits eine Quelle angemeldet."
				);
			}
			else if(state == ClientSenderInterface.STOP_SENDING_NO_RIGHTS) {
				logger.warning("Sendesteuerung fuer Quittung " + object + ", " + dataDescription + " : STOP_SENDING_NO_RIGHTS");
			}
		}
	}
}
