/*
 * Copyright 2019-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.datatree;

import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.ArchiveTimeIndex;
import de.bsvrz.ars.ars.persistence.index.ArchiveTimeIndexImpl;
import de.bsvrz.ars.ars.persistence.index.ContainerManagementIndex;
import de.bsvrz.ars.ars.persistence.index.CorruptIndexException;
import de.bsvrz.ars.ars.persistence.index.DataIndexAndArchiveTimeIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndexImpl;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndex;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndexImpl;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexValues;
import de.bsvrz.ars.ars.persistence.index.backend.management.BaseIndex;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Map;

/**
 * Diese Klasse verwaltet die vom aktuellen Thread geöffneten Indexe. Wenn {@linkplain SyncKey die Synchronisierung} auf eine Datenidentifikation beendet wird, werden automatisch alle Indexe hier freigegeben und können dann von anderen Threads in Anspruch genommen werden.
 *
 * @author Kappich Systemberatung
 */
public final class IndexTree {
	
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Limit ab dem der User über zu viele Containerdateien in einer DID gewarnt wird.
	 * 32000 ist schon recht hoch, es gibt aber viele "volle" Archivsysteme, und wir wollen keinen Debug-Spam erzeugen.
	 */
	public static final long ENTRIES_WARN_LIMIT = Short.MAX_VALUE; // Halbes Dateilimit pro Ordner für FAT32, willkürlich gewählt

	/**
	 * Cache mit offenen Indexen je Thread. ThreadLocal wird hauptsächlich deswegen verwendet, damit die HashMap nicht synchronisiert werden muss (langsam/deadlock-Gefahr).
	 * Das ThreadLocal ist hier <b>nicht</b> static, damit Testfälle mehrere Archivsysteme gleichzeitig starten können und sich die Caches nicht in die Quere kommen.
	 */
	private final ThreadLocal<Map<IndexId, BaseIndex<? extends Enum<?>>>> _cache = ThreadLocal.withInitial(HashMap::new);

	/**
	 * Persistenz
	 */
	private final PersistenceManager _persistenceManager;

	/**
	 * Maximale Anzahl Cache-Einträge für Datenzeitindex.
	 */
	private final int _dataTimeSize;

	/**
	 * Maximale Anzahl Cache-Einträge für Datenindex-Index.
	 */
	private final int _dataIndexSize;

	/**
	 * Maximale Anzahl Cache-Einträge für Archivzeit-Index.
	 */
	private final int _archiveTimeSize;

	/**
	 * Maximale Anzahl Cache-Einträge für Verwaltungsdatenindex.
	 */
	private final int _managementIndexSize;

	/**
	 * Konstruktor.
	 * @param maximumIndexSizeBytes Maximale Anzahl Bytes, die ein Index im Speicher Cachen darf
	 * @param persistenceManager    Verwaltung der Persistenz
	 */
	public IndexTree(final int maximumIndexSizeBytes, final PersistenceManager persistenceManager) {
		_dataTimeSize = Math.min(1, maximumIndexSizeBytes / DataTimeIndexImpl.entrySize());
		_dataIndexSize = Math.min(1, maximumIndexSizeBytes / DataIndexIndexImpl.entrySize());
		_archiveTimeSize = Math.min(1, maximumIndexSizeBytes / ArchiveTimeIndexImpl.entrySize());
		_managementIndexSize = Math.min(1, maximumIndexSizeBytes / ContainerManagementIndex.entrySize());
		_persistenceManager = persistenceManager;
	}

	/**
	 * Schließt alle Indexe, die vom aktuellen Thread geöffnet wurden.
	 */
	public void closeIndexes() {
		Map<IndexId, BaseIndex<? extends Enum<?>>> indexMap = _cache.get();
		for (BaseIndex<? extends Enum<?>> baseIndex : indexMap.values()) {
			try {
				baseIndex.close();
			} catch (IndexException e) {
				_debug.warning("Fehler beim Schließen eines Index", e);
			}
		}
		indexMap.clear();
	}

	/**
	 * Lädt einen Index von der Festplatte. Dies darf nur ausgeführt werden, wenn die Indexdatei noch nicht geöffnet ist. (Eine Indexdatei zweimal zu öffnen kann Zu Datenkorruption führen, da die Inhalte gecacht werden und dann nicht mehr zu den Dateien passen)
	 * @param key   Information, welcher Index geladen werden soll
	 * @param lock  Synchronisierungsschlüssel
	 * @return geladener Index
	 * @throws IndexException Index konnte nicht geladen werden. In manchen Fällen wird die Subklasse {@link CorruptIndexException} geworfen.
	 */
	private BaseIndex<? extends Enum<?>> loadIndexFromDisk(@NotNull IndexId key, SyncKey<IdDataIdentification> lock)
		throws IndexException {
		BaseIndex<? extends Enum<?>> index = loadIndexFromDiskHelper(key, lock);
		if(index instanceof ContainerManagementIndex && index.numEntries() > ENTRIES_WARN_LIMIT) {
			_persistenceManager.warnAboutHugeContainerDirectory(index.getLocation(), index.numEntries());
		}
		return index;
	}

	private BaseIndex<? extends Enum<?>> loadIndexFromDiskHelper(@NotNull final IndexId key, SyncKey<IdDataIdentification> lock) throws IndexException {
		Path indexFile = key.toFile(_persistenceManager);
		if(!Files.exists(indexFile)) {
			// Index fehlt, wird neu aufgebaut
			return rebuildIndex(key, indexFile, lock);
		}
		try {
			return openIndex(key, indexFile);
		}
		catch(CorruptIndexException e) {
			// Index ist fehlerhaft, wird neu aufgebaut
			
			// Zuerst alten Index löschen
			try {
				Files.deleteIfExists(indexFile);
			} catch (IOException ioException) {
				IndexException subExp =
					new IndexException("Defekter Index konnte nicht gelöscht werden", indexFile, ioException);
				subExp.addSuppressed(e);
				throw subExp;
			}
			
			// Neuen Index erstellen und zurückgeben
			return rebuildIndex(key, indexFile, lock);
		}
	}

	/**
	 * Baut einen Index neu auf.
	 * @param indexId   Ort und Typ des Indexes
	 * @param indexFile Anzulegende Indexdatei
	 * @param lock      Synchronisierung auf Datenidentifikation
	 * @return          Neuer Index
	 * @throws IndexException Index konnte nicht angelegt werden
	 */
	private BaseIndex<? extends Enum<?>> rebuildIndex(final IndexId indexId, final Path indexFile, SyncKey<IdDataIdentification> lock) throws IndexException {
		if(indexId.getIndexClass() == IndexImpl.ManagementData) {
			return rebuildContainerHeaderIndex(indexId, indexFile, null);
		}
		return rebuildStandardIndex(indexId, indexFile, lock);
	}

	/**
	 * Baut einen Standard-Index (Datenzeit, Datenindex, Archivzeit) neu auf.
	 * @param indexId   Ort und Typ des Indexes
	 * @param indexFile Anzulegende Indexdatei
	 * @param lock      Synchronisierung auf Datenidentifikation
	 * @return          Neuer Index
	 * @throws IndexException Index konnte nicht angelegt werden
	 */
	private BaseIndex<IndexValues> rebuildStandardIndex(final IndexId indexId, final Path indexFile, SyncKey<IdDataIdentification> lock) throws IndexException {
		// Ein Standardindex wird aus dem Verwaltungsdatenindex erzeugt.
		
		IndexId managementId = new IndexId(indexId.getContainerFileDir(), IndexImpl.ManagementData);
		ContainerManagementIndex managementIndex = cacheIndex(managementId, loadIndexFromDisk(managementId, lock));
		return rebuildStandardIndex(indexId, indexFile, managementIndex);
	}

	/**
	 * Baut einen Standard-Index (Datenzeit, Datenindex, Archivzeit) aus einem Verwaltungsdatenindex neu auf.
	 *
	 * @param indexId         Ort und Typ des Indexes
	 * @param indexFile       Anzulegende Indexdatei
	 * @param managementIndex Geladener Verwaltugnsdatenindex
	 *
	 * @return Neuer Index
	 *
	 * @throws IndexException Index konnte nicht angelegt werden
	 */
	private BaseIndex<IndexValues> rebuildStandardIndex(IndexId indexId, Path indexFile, ContainerManagementIndex managementIndex) throws IndexException {
		deleteIndex(indexId);
		BaseIndex<IndexValues> index = cacheIndex(indexId, openIndex(indexId, indexFile));
		_persistenceManager.rebuildStandardIndex(managementIndex, indexId.getContainerFileDir(), index);
		return index;
	}

	/**
	 * Baut einen Verwaltungsdatenindex neu auf.
	 *
	 * @param indexId       Ort und Typ des Indexes
	 * @param indexFile     Anzulegende Indexdatei
	 * @param rebuildResult Hier kann die Methode eine Statistik über Vorgang ablegen. Kann null sein.
	 * @return Neuer Index
	 *
	 * @throws IndexException Index konnte nicht angelegt werden
	 */
	private ContainerManagementIndex rebuildContainerHeaderIndex(final IndexId indexId, final Path indexFile,
	                                                             @Nullable final PersistenceManager.RebuildResult rebuildResult) throws IndexException {
		deleteIndex(indexId);
		ContainerManagementIndex index = cacheIndex(indexId, openIndex(indexId, indexFile));
		
		// Implementierung in Verwaltungsschicht, da Informationen aus Containern geholt werden.
		_persistenceManager.rebuildContainerHeaderIndex(indexId.getContainerFileDir(), index, rebuildResult);
		
		return index;
	}

	/**
	 * Entfernt einen Index umgehend aus dem Cache und aktualisiert die Indexdatei auf der Festplatte auf de aktuellen gecachten Stand.
	 * @param indexId Index-Typ und -Ort
	 * @throws IndexException Bei einem Schreibfehler
	 */
	private void invalidateNow(final IndexId indexId) throws IndexException {
		BaseIndex<? extends Enum<?>> removed = _cache.get().remove(indexId);
		if(removed != null) {
			removed.close();
		}
	}

	/**
	 * Öffnet einen Index von der Festplatte.
	 * @param indexId   Typ und Ort des Index
	 * @param indexFile Indexdatei
	 * @return Geladener Index
	 * @throws CorruptIndexException Falls beim Laden ein Fehler auftritt. Dann ist die Indexdatei vermutlich defekt und muss neu aufgebaut werden.
	 */
	private BaseIndex<? extends Enum<?>> openIndex(final IndexId indexId, final Path indexFile) throws CorruptIndexException {
		switch(indexId.getIndexClass()) {
			case DataTime:
				return new DataTimeIndexImpl(_dataTimeSize, indexFile);
			case DataIndex:
				// DataIndexIndex alleinstehend nur bei nachgeforderten Daten, bei Online-Daten wird der Archivzeitindex mitbenutzt,
				// der dort auch im Datenindex monoton ist.
				assert indexId.getContainerFileDir().getArchiveDataKind().isRequested();
				return new DataIndexIndexImpl(_dataIndexSize, indexFile);
			case ArchiveTime:
				if(indexId.getContainerFileDir().getArchiveDataKind().isRequested()) {
					return new ArchiveTimeIndexImpl(_archiveTimeSize, indexFile);
				}
				else {
					return new DataIndexAndArchiveTimeIndex(_archiveTimeSize, indexFile);
				}
			case ManagementData:
				return new ContainerManagementIndex(_managementIndexSize, indexFile);
		}
		throw new IllegalArgumentException(String.valueOf(indexId.getIndexClass()));
	}

	public ContainerManagementIndex getContainerManagementIndex(SyncKey<IdDataIdentification> lock, ArchiveDataKind archiveDataKind) throws IndexException {
		if(!lock.isValid()) {
			throw new IllegalArgumentException("Der Parameter \"lock\" ist ungültig:" + lock);
		}
		IndexId key = new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.ManagementData);
		return cacheIndex(key, loadIndexFromDisk(key, lock));
	}
	
	public DataIndexIndex getDataIndexIndex(SyncKey<IdDataIdentification> lock, ArchiveDataKind archiveDataKind) throws IndexException {
		if(!lock.isValid()) {
			throw new IllegalArgumentException("Der Parameter \"lock\" ist ungültig:" + lock);
		}
		if(!archiveDataKind.isRequested()){
			return (DataIndexIndex)getArchiveTimeIndex(lock, archiveDataKind);
		}
		IndexId key = new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.DataIndex);
		return cacheIndex(key, loadIndexFromDisk(key, lock));
	}

	public DataTimeIndex getDataTimeIndex(SyncKey<IdDataIdentification> lock, ArchiveDataKind archiveDataKind) throws IndexException {
		if(!lock.isValid()) {
			throw new IllegalArgumentException("Der Parameter \"lock\" ist ungültig:" + lock);
		}
		IndexId key = new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.DataTime);
		return cacheIndex(key, loadIndexFromDisk(key, lock));
	}	
	
	public ArchiveTimeIndex getArchiveTimeIndex(SyncKey<IdDataIdentification> lock, ArchiveDataKind archiveDataKind) throws IndexException {
		if(!lock.isValid()) {
			throw new IllegalArgumentException("Der Parameter \"lock\" ist ungültig:" + lock);
		}
		IndexId key = new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.ArchiveTime);
		return cacheIndex(key, loadIndexFromDisk(key, lock));
	}

	@SuppressWarnings("unchecked")
	private <T extends BaseIndex<?>> T cacheIndex(IndexId key, BaseIndex<? extends Enum<?>> indexToCache) {
		Map<IndexId, BaseIndex<? extends Enum<?>>> map = _cache.get();
		BaseIndex<? extends Enum<?>> index = map.get(key);
		if (index != null) {
			return (T) index;
		}
		
		// Nicht computeIfAbsent verwenden, da loadIndex ggf. weitere andere Indexe lädt und es sonst zu einer
		// java.util.ConcurrentModificationException kommt 
		if(map.containsKey(key)) {
			throw new IllegalArgumentException("key mehrfach initialisiert");
		}
		map.put(key, indexToCache);
		return (T) indexToCache;
	}

	/**
	 * Erstellt alle Indexe aus den Containerdaten neu.
	 * @param lock            Synchronisierung auf Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 * @param result          Ergebnis (Statistik) oder null
	 * @throws IndexException Falls die Indexe nicht neu erstellt werden konnten
	 */
	public void recreateIndex(SyncKey<IdDataIdentification> lock, ArchiveDataKind archiveDataKind, @Nullable PersistenceManager.RebuildResult result) throws IndexException {
		ContainerManagementIndex managementIndex = rebuildContainerHeaderIndex(lock, archiveDataKind, result);
		if(archiveDataKind.isRequested()) {
			rebuildStandardIndex(lock, archiveDataKind, IndexImpl.DataIndex, managementIndex);
		}
		rebuildStandardIndex(lock, archiveDataKind, IndexImpl.ArchiveTime, managementIndex);
		rebuildStandardIndex(lock, archiveDataKind, IndexImpl.DataTime, managementIndex);
	}

	/**
	 * Erstellt den Verwaltungsdatenindex neu.
	 * @param lock            Synchronisierung auf Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 * @param rebuildResult   Ergebnis (Statistik)
	 * @return Neuer Verwaltugnsdatenindex
	 * @throws IndexException Falls der Index nicht erstellt werden konnte
	 */
	private ContainerManagementIndex rebuildContainerHeaderIndex(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind, @Nullable PersistenceManager.RebuildResult rebuildResult) throws IndexException {
		IdContainerFileDir containerFileDir = new IdContainerFileDir(lock.getElement(), archiveDataKind);
		IndexId indexId = new IndexId(containerFileDir, IndexImpl.ManagementData);
		return rebuildContainerHeaderIndex(indexId, indexId.toFile(_persistenceManager), rebuildResult);
	}

	/**
	 * Erstellt einen Standard-Index aus dem Verwaltungsdatenindex neu.
	 * @param lock            Synchronisierung auf Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 * @param indexClass Art des zu erstellenden Index
	 * @param managementIndex Verwaltungsdatenindex
	 * @return Neuer Index
	 * @throws IndexException Falls der Index nicht erstellt werden konnte
	 */
	@SuppressWarnings("UnusedReturnValue")
	private BaseIndex<IndexValues> rebuildStandardIndex(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind, final IndexImpl indexClass,
	                                                    ContainerManagementIndex managementIndex) throws IndexException {
		IdContainerFileDir containerFileDir = new IdContainerFileDir(lock.getElement(), archiveDataKind);
		IndexId indexId = new IndexId(containerFileDir, indexClass);
		return rebuildStandardIndex(indexId, indexId.toFile(_persistenceManager), managementIndex);
	}

	/**
	 * Löscht alle Indexe und Indexdateien in einem Containerverzeichnis.
	 * @param lock            Synchronisierung auf Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 * @param result          Statistik (die Zahl der gelöschten Indexdateien wird in result.containerIgnored gespeichert)
	 * @throws IndexException Fehler beim Löschen
	 */
	public void deleteIndex(SyncKey<IdDataIdentification> lock, ArchiveDataKind archiveDataKind, @Nullable PersistenceManager.RebuildResult result) throws IndexException {
		int sum = 0;
		sum += deleteIndex(new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.ManagementData));
		if(archiveDataKind.isRequested()) {
			sum += deleteIndex(new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.DataIndex));
		}
		sum += deleteIndex(new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.ArchiveTime));
		sum += deleteIndex(new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), IndexImpl.DataTime));
		if(result != null) {
			result.containerIgnored += sum;
		}
	}

	/**
	 * Löscht einen Index.
	 * @param indexId ID des Indexes    
	 * @return Anzahl gelöschter Dateien (1 = Index gelöscht, 0 = Index war nicht vorhanden)
	 * @throws IndexException Wenn der Index nicht gelöscht werden konnte (in der Regel Dateisystemproblem)
	 */
	private int deleteIndex(final IndexId indexId) throws IndexException {
		invalidateNow(indexId);
		Path file = indexId.toFile(_persistenceManager);
		try {
			return Files.deleteIfExists(file) ? 1 : 0;
		} catch (IOException e) {
			throw new IndexException("Kann vorhandenen Index nicht löschen", file, e);
		}
	}

	/**
	 * Schließt alle vom aktuellen Thread geöffneten Indexe und schreibt die Puffer auf die Festplatte.
	 * @param lock            Synchronisierung auf Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 * @throws IndexException Fehler beim Schreiben der Indexdateien
	 */
	public void closeIndexes(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind) throws IndexException {
		for(IndexImpl value : IndexImpl.values()) {
			IndexId id = new IndexId(new IdContainerFileDir(lock.getElement(), archiveDataKind), value);
			invalidateNow(id);
		}
	}

	/**
	 * Schreibt die Puffer für alle vom aktuellen Thread geöffneten Indexe auf die Festplatte ohne den Index zu schließen.
	 * Fehler werden dabei ignoriert udn führen zu einer Warnung.
	 * Diese Methode wird vom Datenkonsistenten Backup benutzt. Hier ist bei Fehlern keine bessere Fehlerbehandlung sinnvoll.
	 * 
	 * @param lock            Synchronisierung auf Datenidentifikation
	 * @param archiveDataKind Archivdatenart
	 */
	public void flushIndexes(final SyncKey<IdDataIdentification> lock, final ArchiveDataKind archiveDataKind) {
		for (Map.Entry<IndexId, BaseIndex<? extends Enum<?>>> entry : _cache.get().entrySet()) {
			IdContainerFileDir containerFileDir = entry.getKey().getContainerFileDir();
			if(containerFileDir.getDataIdentification().equals(lock.getElement()) && containerFileDir.getArchiveDataKind() == archiveDataKind) {
				BaseIndex<? extends Enum<?>> baseIndex = entry.getValue();
				try {
					baseIndex.flush();
				} catch (IndexException e) {
					_debug.warning("Fehler beim Speichern eines Index", e);
				}
			}
		}
	}

	/**
	 * Debug-Methode. Wirft eine Exception wenn der aktuelle Thread aktuell Idnexe geöffnet hat.
	 */
	public void ensureNoCached() {
		if(!_cache.get().isEmpty()) {
			throw new IllegalStateException("Cache nicht leer: " + _cache.get());
		}
	}
}
