/*
 * Copyright 2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * TBD Dokumentation
 *
 * @author Kappich Systemberatung
 */
public abstract class AbstractTask {
	/**
	 * Logger
	 */
	protected static final Debug _debug = Debug.getLogger();
	/**
	 * Zeigt an ob der Task beendet werden soll. Der Task bricht dann möglichst schnell an der nächsten sicheren Stelle ab.
	 */
	protected volatile boolean _shouldTerminate;

	/** Datenindentifikationen */
	protected DataIdentTree didTree;

	/** Zentrale Archiv-Verwaltung. */
	protected final ArchiveManager archMgr;

	/** PersistenceManager für den Zugriff auf die Daten des Archivsystems */
	protected PersistenceManager persMgr;

	/**
	 * Erstellt einen neuen Task
	 *
	 * @param archiveMgr Archiv-Manager (Kann für Testfälle und ähnliches null sein, allerdings kann es dann passieren, das bestimmte
	 *                   Funktionen nicht funktionieren.)
	 */
	public AbstractTask(ArchiveManager archiveMgr) {
		archMgr = archiveMgr;
		if(archMgr != null) {
			didTree = archMgr.getDataIdentTree();
			persMgr = archMgr.getPersMgr();
			if(archiveMgr.getDavCon() != null) {
				getObjectsFromDav();
			}
		}
	}
	
	public abstract void start();

	/**
	 * Terminiert den Task indem dafür gesorgt wird, das {@link #shouldTerminate()} true zurückliefert. Der Task
	 * sollte diese Funktion regelmäßig aufrufen um dann bei Rückgabe von {@code true} den Task zeitnah zu beenden.
	 */
	public void terminateTask() {
		_shouldTerminate = true;
	}

	/**
	 * Zeigt an, ob der Task beendet werden soll.
	 *
	 * @return {@code true} falls der Task beendet werden soll.
	 */
	public boolean shouldTerminate() {
		return _shouldTerminate;
	}

	public ArchiveManager getArchMgr() {
		return archMgr;
	}

	/**
	 * Hält den Task an, falls sich das Archivsystem in einem kritischen Zustand befindet. Wird das Archivsystem beendet, wird der Task fortgesetzt, so dass er
	 * seine Aufgabe noch vollständig erfüllen kann. {@link #shouldTerminate()} liefert anschließend {@code true} zurück.
	 */
	public final void suspendTaskIfNecessary() {
		try {
			ArchiveManager archiveManager = getArchMgr();
			if(archiveManager != null) archiveManager.getInQueuesMgr().suspendTaskIfNecessary(this);
		}
		catch(InterruptedException e) {
			AbstractTask._debug.warning("Thread wurde beim Warten unterbrochen", e);
		}
	}

	/**
	 * Wird aufgerufen, sobald eine Datenverteilerverbindung verfügbar ist. Kann benutzt werden, um Objekte zu laden und Anmeldungen durchzuführen.
	 * <p>
	 * Wichtig: Diese Methode wird unter Umständen im Konstruktor ausgeführt und sollte daher nicht erwarten, dass die Subklasse (`this`) vollständig initialisiert ist.
	 */
	public void getObjectsFromDav() {
	}

	public abstract String getName();

	public abstract void setName(String name);

	public abstract void join() throws InterruptedException;

	public abstract boolean isAlive();

	public abstract boolean isTerminated();
}
