/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.concurrent.Semaphore;

/**
 * Task zur Archivierung nachgefordeter Daten. Unterscheidet sich hauptsaechlich in der Behandlung von Datenlücken von {@link ArchiveTask}.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveRequestedTask extends ArchiveTask {

	private ArchiveData ad;

	/**
	 * Erzeugt einen ArchiveRequestedTask
	 *
	 * @param archiveMgr Archiv-Manager
	 * @param tskIndex   Index des Tasks
	 * @param inQueue    Eingangs-Warteschlange
	 */
	public ArchiveRequestedTask(ArchiveManager archiveMgr,int tskIndex, ArchiveRingBuffer inQueue) {
		super(archiveMgr, tskIndex, inQueue);
	}

	/**
	 * Alle Instanzvariablen vorbelegen.
	 *
	 * @param archiveData Archivdatensatz
	 * @param archTime    Archivzeitstempel
	 */
	private void presetData(ArchiveData archiveData, long archTime) {
		ad = archiveData;
		arsTime = archTime;
		dataTime = ad.getDataTime();
		dataIdx = ad.getDataIndex();
		adk = ad.getDataKind().isDelayed() ? ArchiveDataKind.REQUESTED_DELAYED : ArchiveDataKind.REQUESTED;
		din = didTree.get(archiveData);
		didPath = new DataIdentificationDir(persMgr, din);
	}

	@Override
	public void work(Object archiveData, long archTime) {
		if(archiveData instanceof Semaphore) {
			// RequestGapTask aufwecken
			Semaphore semaphore = (Semaphore) archiveData;
			semaphore.release();
			return;
		}
		logReceivedData((ArchiveData)archiveData);
		try {
			// Flags, IDs, Variable besetzen, DataIdentNode holen, etc.:
			presetData((ArchiveData) archiveData, archTime);

			IdDataIdentification dataIdentification = din.getDataIdentification();
			
			lookupLastDataIdxes(dataIdentification, true);        // Letzte Datenindexe, Datenzeiten nachsehen

			openContID = persMgr.getOpenContID(new IdContainerFileDir(dataIdentification, adk));
			dataEntryLen = _serializer.serializeReqData(ad, arsTime, compresser);
			(storeData() ? _successCount : _failCount).incrementAndGet();
		}
		catch(IndexException | PersistenceException | SynchronizationFailedException e) {
			handleErr(e, Util.ad2Str(ad));
		}
	}

	protected void logReceivedData(ArchiveData ad) {
		_debug.finer(getName() + ": nachgeforderten Datensatz zur Archivierung erhalten: ", Util.ad2Str(ad));
	}

}
