/*
 * Copyright 2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import com.google.common.base.Joiner;
import com.google.common.collect.Multimap;
import com.google.common.collect.MultimapBuilder;
import com.google.common.collect.Multimaps;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.StartupProperties;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.util.FileSync;
import de.bsvrz.ars.ars.persistence.util.SyncResult;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirAction;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirWalk;
import de.bsvrz.ars.ars.persistence.walk.internal.StatusPrinter;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.util.cron.CronDefinition;
import de.bsvrz.dav.daf.util.cron.CronScheduler;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import java.io.IOException;
import java.io.InputStream;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.channels.OverlappingFileLockException;
import java.nio.file.Files;
import java.nio.file.InvalidPathException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * {@link AbstractTask}-Implementierung vom datenkonsistenten Backup.
 * 
 * <p>Grundgedanke des datenkonsitenten Backups ist, dass es ausreichend ist, alle Containerverzeichnisse (bzw. Datenidentifikationen) einzeln in einem konsistenten Zustand zu sichern.</p>
 * <p>Zur Umsetzung dieses Konzepts würde im Archivsystem ein weiterer Hintergrundtask integriert werden, der alle Containerverzeichnisse nacheinander durchgeht und einzeln sichert. Eine parallele Bearbeitung von mehreren Verzeichnissen ist prinzipiell problemlos möglich, könnte jedoch die Systemlast erhöhen und sollte daher nicht standardmäßig aktiviert werden.
 * Der neue Task wird regelmäßig folgende Aktionen ausführen:</p>
 * <ul>
 * <li>Für jedes Containerdateienverzeichnis:<ol>
 * <li>Schreibe die Indexe, leere den Cache und unterdrücke weitere Schreibvorgänge</li>
 * <li>Stelle sicher, dass das entsprechende Backup-Verzeichnis auf demselben Stand ist, wie das aktuelle Archivdatenverzeichnis im Persistenzverzeichnis. Das entsprechende Unterverzeichnis im Backup-Ordner wird angelegt, wenn es noch nicht existiert. </li>
 * <li>Daten, die im Backup-Verzeichnis noch nicht existieren, werden kopiert, Daten, die auf einem alten bzw. abweichenden Stand sind, werden aktualisiert und Daten, die nicht mehr existieren (z. B. weil sie aus dem Archiv ausgelagert wurden), werden im Backup-Verzeichnis ebenfalls gelöscht.</li>
 * <li>Hebe die Blockierung wieder auf </li>
 * </ol>
 * </li>
 * <li>Schreibe am Ende eine _startUpInfo.property und _restartTime.property, die den Stand der gesicherten Daten angibt
 * (die Inhalte können bei der Sicherung aggregiert werden).</li>
 * <li>Starte (auf Wunsch) einen vom Anwender konfigurierbaren Prozess, der die Daten aus diesem Verzeichnis noch einmal an einen anderen Ort sichert oder beispielsweise einen Schnappschuss erstellt.</li>
 * <li>Warte ggf. auf die Beendigung dieses Prozesses</li>
 * </ul>
 *
 * @author Kappich Systemberatung
 */
public final class MirrorTask extends CronTask {

	private static final Debug _debug = Debug.getLogger();
	private static final String[] BYTE_UNIT = {" Byte", " KiB", " MiB", " GiB", " TiB", " PiB", " EiB"};

	/**
	 * {@link CronScheduler} zum regelmäßigen Aktualisieren der Definitionsdateien von der Festplatte 
	 */
	private final CronScheduler _refreshScheduler = new CronScheduler();
	
	private final LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;

	/**
	 * Konstruktor
	 *
	 * @param archiveMgr Archivmanager
	 * @param longTermTaskStatePublisher
	 */
	public MirrorTask(final ArchiveManager archiveMgr, final LongTermTaskStatePublisher longTermTaskStatePublisher) {
		super(archiveMgr);
		setName("Datenkonsistentes Backup");
		_refreshScheduler.schedule(this::refreshCronTasks, CronDefinition.EVERY_HOUR);
		refreshCronTasks();
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Datenkonsistentes Backup");
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
	}

	@Override
	public void start() {
		super.start();
	}

	@Override
	public void terminateTask() {
		super.terminateTask();
		_refreshScheduler.shutdownNow();
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
	}

	private void refreshCronTasks() {
		try {
			Path backupDir = archMgr.getPersMgr().getArchivePath().resolve("backup");
			Files.createDirectories(backupDir);
			List<MirrorDefinition> definitions = new ArrayList<>();
			try(Stream<Path> list = Files.list(backupDir)) {
				list.forEach(new Consumer<Path>() {
					@Override
					public void accept(final Path path) {
						if(path.getFileName().toString().toLowerCase(Locale.ENGLISH).endsWith(".properties")) {
							Properties properties = new Properties();
							try(InputStream inStream = Files.newInputStream(path)) {
								properties.load(inStream);
							}
							catch(IOException e) {
								_debug.warning("Datei " + path + " konnte nicht gelesen werden.", e);
								return;
							}
							String cronString = properties.getProperty("cron");
							String targetString = properties.getProperty("ziel");
							if(cronString == null) {
								_debug.warning("Datei " + path + " konnte nicht gelesen werden: Sie hat keinen Parameter \"cron\".");
								return;
							}
							if(targetString == null) {
								_debug.warning("Datei " + path + " konnte nicht gelesen werden: Sie hat keinen Parameter \"ziel\".");
								return;
							}
							
							Path targetBaseDir;
							try {
								targetBaseDir = Paths.get(targetString.trim());
							}
							catch (InvalidPathException e) {
								_debug.warning("Das in Datei \"" + path + "\" angegebene Ziel \"" + targetString + "\" ist kein gültiger Pfadname. Die Backupdefinition wird ignoriert.", e);
								return;
							}

							String commandString = properties.getProperty("cmd");
							String[] commandLine = getCommand(commandString);
							
							CronDefinition cronDefinition;
							try {
								cronDefinition = new CronDefinition(cronString);
							}
							catch (IllegalArgumentException e) {
								_debug.warning("Der in Datei \"" + path + "\" angegebene Cron-String \"" + cronString + "\" ist ungültig. Die Backupdefinition wird ignoriert.", e);
								return;
							}
							
							MirrorDefinition e = new MirrorDefinition(cronDefinition, targetBaseDir, commandLine);
							definitions.add(e);
						}
					}

					private String[] getCommand(final String command) {
						if(System.getProperty("os.name", "unknown").startsWith("Windows")) {
							return new String[]{"cmd", "/C", command};
						}
						return new String[]{"sh", "-c", command};
					}
				});
			}
			setCronTaskDefinitions(definitions.stream().map(this::createMirrorCronTask).collect(Collectors.toList()));
		}
		catch(IOException e) {
			_debug.warning("Kann Backup-Definitionen nicht einlesen", e);
		}
	}

	/**
	 * Erstellt aus einer {@link MirrorDefinition} eine {@link de.bsvrz.ars.ars.mgmt.tasks.CronTask.CronTaskDefinition}
	 * @param mirrorDefinition Enthält Einstellungen aus einer Einstellungsdatei (backup/???.properties)
	 * @return CronTaskDefinition, die an den {@link de.bsvrz.dav.daf.util.cron.CronScheduler} übergeben werden kann
	 */
	private CronTaskDefinition createMirrorCronTask(final MirrorDefinition mirrorDefinition) {
		return new CronTaskDefinition(mirrorDefinition.getCronDefinition(), () -> doMirror(mirrorDefinition.getTargetDirectory(), mirrorDefinition.getCommandLine()));
	}

	/**
	 * Führt einen Backup-lauf durch
	 * @param targetBaseDir Zielverzeichnis
	 * @param commandLine   Kommando, das nach dem erfolgreichen Ausführen ausgeführt wird.
	 */
	public void doMirror(final Path targetBaseDir, final String[] commandLine) {
		if(shouldTerminate()) {
			return;
		}
		try {
			Thread.currentThread().setName("Datenkonsistentes Backup nach " + targetBaseDir);
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);
			Files.createDirectories(targetBaseDir);
			Path targetFlagFile = targetBaseDir.resolve(PersistenceManager.ACTIVITY_FLAG_FILE_NAME);
			Path startUpInfoFile = targetBaseDir.resolve(StartupProperties.STARTUP_PROP_FILE_NAME);
			Path restartTimeFile = targetBaseDir.resolve(PersistenceManager.RESTART_TIME_FILE_NAME);
			
			Files.deleteIfExists(startUpInfoFile);
			Files.deleteIfExists(restartTimeFile);
			
			MirrorAction action = new MirrorAction(targetBaseDir);
			try(FileChannel channel = FileChannel.open(targetFlagFile, StandardOpenOption.CREATE, StandardOpenOption.DELETE_ON_CLOSE, StandardOpenOption.WRITE)) {
				FileLock fileLock = channel.tryLock();
				if(fileLock == null) {
					_debug.warning("Ein anderes datenkonsistentes Backup ist bereits im Verzeichnis " + targetBaseDir + " aktiv");
					return;
				}
				try {
					_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
					DataIdentificationDirWalk walk = DataIdentificationDirWalk.allDirectories(persMgr);
					walk.setPublisherTask(_longTermTaskStatePublisherTask);
					walk.setStatusPrinter(new StatusPrinter() {
						@Override
						public String getStatusMessage(final String actionName, final Duration runtime, final boolean indexFinished, final long indexedContainerDirs, final long visitedContainerDirs, final long visitedContainerFiles) {
							return super.getStatusMessage(actionName, runtime, indexFinished, indexedContainerDirs, visitedContainerDirs, visitedContainerFiles) +
									getMessage(action, runtime);
						}

						@Override
						public String getSuccessMessage(final String actionName, final Duration runtime, final long visitedContainerDirs, final long visitedContainerFiles) {
							return super.getSuccessMessage(actionName, runtime, visitedContainerDirs, visitedContainerFiles) +
									getMessage(action, runtime);
						}
					});
					walk.execute("Datenkonsistentes Backup nach " + targetBaseDir, 1, action);
					new FileSync().syncDirectories(persMgr.getArchivePath().resolve("backup"), targetBaseDir.resolve("backup"));
					_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
					if(shouldTerminate()) {
						return;
					}
					if(action.isSuccessful()) {
						writeStartUpInfo(action, targetBaseDir);
						writeRestartTime(action, targetBaseDir);
					}
				}
				finally {
					fileLock.release();
				}
			}
			catch(OverlappingFileLockException e) {
				_debug.warning("Ein anderes datenkonsistentes Backup ist bereits im Verzeichnis " + targetBaseDir + " aktiv");
				return;
			}
			if(commandLine.length > 0) {
				runExternalProcess(targetBaseDir, commandLine);
			}
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
		catch(PersistenceException | IOException | InterruptedException e) {
			_debug.error("Unerwarteter Fehler beim konsistenten Backup", e);
		}
	}

	/**
	 * Formatiert eine Debug-Meldung über den Fortschritt
	 * @param action  Aktuelle Spiegelungs-Aktion (enthält Statisktik-Daten)
	 * @param runtime Bisherige Laufzeit
	 * @return Formatierter String
	 */
	@NotNull
	private static String getMessage(final MirrorAction action, final Duration runtime) {
		String byteSize = formatByteSize(action.getNumBytesCopied());
		String speed = formatSpeed(action.getNumBytesCopied() / (runtime.getSeconds() + runtime.getNano() / 1000000000.0));
		return "\n " + getFileCountString(action.getNumFilesCopied()) + " (" + byteSize + "; " + speed + ") kopiert."
				+ "\n " + getFileCountString(action.getNumFilesDeleted()) + " gelöscht.";
	}

	/**
	 * Formatiert einen Byte-Pro-Sekunde-Wert
	 * @param bytesPerSecond Byte pro Sekunde
	 * @return String-Darstellung
	 */
	private static String formatSpeed(final double bytesPerSecond) {
		if(!Double.isFinite(bytesPerSecond)) return "?";

		double bytes_double = bytesPerSecond;
		for(final String unit : BYTE_UNIT) {
			if(bytes_double < 1024L) {
				return String.format("%.1f", bytes_double) + unit + "/s";
			}
			bytes_double /= 1024;
		}
		return bytesPerSecond + BYTE_UNIT[0] + "/s";	
	}

	/**
	 * Formatiert eine Datenmenge in Bytes
	 * @param bytes Bytes
	 * @return String-Darstellung
	 */
	@NotNull
	private static String formatByteSize(final long bytes) {
		double bytes_double = bytes;
		for(final String unit : BYTE_UNIT) {
			if(bytes_double < 1024L) {
				return String.format("%.1f", bytes_double) + unit;
			}
			bytes_double /= 1024;
		}
		return bytes + BYTE_UNIT[0];
	}

	/**
	 * Formatiert eine Dateianzahl
	 * @param numFiles Anzahl dateien (>= 0)
	 * @return einen Wert wie "3 Dateien"
	 */
	@NotNull
	private static String getFileCountString(final long numFiles) {
		if(numFiles == 0) {
			return "Keine Dateien";
		}
		else if(numFiles == 1) {
			return "Eine Datei";
		}
		return numFiles + " Dateien";
	}

	/**
	 * Schreibt eine StartUpInfo.properties-Datei ins Spiegelungsverzeichnis. Die Werte werden aus dem aktuellen Mirror-Lauf aggregiert
	 * @param action        Mirrorlauf
	 * @param targetBaseDir Zielverzeichnis (wurzel)
	 * @throws PersistenceException Bei Schreibfehlern
	 */
	private void writeStartUpInfo(final MirrorAction action, final Path targetBaseDir) throws PersistenceException {
		StartupProperties startupProperties = new StartupProperties(targetBaseDir);
		startupProperties.setVal(StartupProperties.STUP_LAST_ATIME, action.getLastATime());
		startupProperties.setVal(StartupProperties.STUP_LAST_USED_MEDIUM_ID, action.getLastMediumId());
		startupProperties.setVal(StartupProperties.STUP_MAX_CONT_ID, action.getMaxContainerId());
		startupProperties.writeStartUpProperties();
	}

	/**
	 * Schreibt die Datei, die enthält bis zu welchem Archivzeitpunkt die angegebenen Daten gültig waren (um beim Neustart Datenlücken korrekt zu bilden)
	 * @param action        Mirrorlauf
	 * @param targetBaseDir Zielverzeichnis (wurzel)
	 */
	private void writeRestartTime(final MirrorAction action, final Path targetBaseDir) {
		persMgr.saveUnsubscriptionTime(action.getArchiveTimes(), targetBaseDir.resolve(PersistenceManager.RESTART_TIME_FILE_NAME));
	}

	/**
	 * Führt einen externen Prozess aus
	 * @param targetBaseDir Zielverzeichnis
	 * @param commandLine   Kommandozeile
	 * @throws IOException  IO-Fehler
	 * @throws InterruptedException Interrupted beim Warten auf Beendigung
	 */
	private static void runExternalProcess(final Path targetBaseDir, final String[] commandLine) throws IOException, InterruptedException {
		
		_debug.info("Führe Befehl aus: '" + Joiner.on("' '").join(commandLine) + "'");
		
		ProcessBuilder processBuilder = new ProcessBuilder().command(commandLine);
		processBuilder.directory(targetBaseDir.toFile());
		processBuilder.inheritIO();
		Process process = processBuilder.start();
		process.waitFor();
	}

	/**
	 * Eigentliche Implementierung der Spiegelungs-Aktion
	 */
	private final class MirrorAction implements DataIdentificationDirAction {

		/** Zielverzeichnis */
		private final Path _targetBaseDir;

		/** Statistik */
		private final AtomicLong _lastATime = new AtomicLong();

		/** Statistik */
		private final AtomicLong _lastMediumId = new AtomicLong();

		/** Statistik */
		private final AtomicLong _maxContainerId = new AtomicLong();

		/** Statistik */
		private final AtomicLong _numFilesDeleted = new AtomicLong();

		/** Statistik */
		private final AtomicLong _numBytesCopied = new AtomicLong();

		/** Statistik */
		private final AtomicLong _numFilesCopied = new AtomicLong();

		/** Wird auf false gesetzt, falls ein Fehler auftrat */
		private boolean _successful = true;

		private final Multimap<Long, IdDataIdentification> _archiveTimes = Multimaps.synchronizedMultimap(MultimapBuilder.treeKeys().arrayListValues().build());

		/**
		 * Erzeugt einen eue Instanz
		 * @param targetBaseDir Zielverzeichnis
		 */
		private MirrorAction(final Path targetBaseDir) {
			this._targetBaseDir = targetBaseDir;
		}

		/**
		 * Wird fü jedes Datenientifikationsverzeichnis ausgeführt
		 * @param dataIdentificationDir DataIdentificationDir
		 * @param walk Aktueller Durchlauf
		 */
		@Override
		public void run(final DataIdentificationDir dataIdentificationDir, final DataIdentificationDirWalk walk) {
			if(shouldTerminate()) {
				_successful = false;
				walk.terminate();
				return;
			}
			suspendTaskIfNecessary();
			_archiveTimes.put(System.currentTimeMillis(), dataIdentificationDir.getDataIdentification());
			try(SyncKey<IdDataIdentification> indexLock = archMgr.lockIndex(dataIdentificationDir.getDataIdentification())) {
				for(ArchiveDataKind archiveDataKind : dataIdentificationDir.getArchiveDataKinds()) {
					new MirrorDirectoryAction(dataIdentificationDir, indexLock, archiveDataKind).execute();
				}
			}
			catch(IOException e) {
				_successful = false;
				_debug.warning("Verzeichnis " + dataIdentificationDir + " konnte nicht vollständig gesichert werden. Fehler beim Kopieren der Dateien", e);
			}
			catch(Exception e) {
				_successful = false;
				_debug.warning("Verzeichnis " + dataIdentificationDir + " konnte nicht vollständig gesichert werden.", e);
			}
		}

		public long getLastATime() {
			return _lastATime.get();
		}

		public long getLastMediumId() {
			return _lastMediumId.get();
		}

		public long getMaxContainerId() {
			return _maxContainerId.get();
		}

		public Multimap<Long, IdDataIdentification> getArchiveTimes() {
			return _archiveTimes;
		}

		public long getNumFilesCopied() {
			return _numFilesCopied.get();
		}
		
		public long getNumFilesDeleted() {
			return _numFilesDeleted.get();
		}

		public long getNumBytesCopied() {
			return _numBytesCopied.get();
		}

		public boolean isSuccessful() {
			return _successful;
		}

		/**
		 * Aktion für ein Containerdateienverzeichnis, hier findet die eigentliche Kopier-Aktion statt
		 */
		private class MirrorDirectoryAction {
			private final IdContainerFileDir _containerFileDir;
			private final SyncKey<IdDataIdentification> _indexLock;

			/**
			 * Konstruktor
			 * @param dataIdentificationDir Verzeichnsi der Datenidentifikation
			 * @param indexLock             Lock auf den Index
			 * @param adk
			 */
			public MirrorDirectoryAction(final DataIdentificationDir dataIdentificationDir, final SyncKey<IdDataIdentification> indexLock,
			                             final ArchiveDataKind adk) {
				_indexLock = indexLock;
				_containerFileDir = new IdContainerFileDir(dataIdentificationDir.getDataIdentification(), adk);
			}

			/**
			 * Führt das aktualisieren und kopieren der daten durch
			 * @throws Exception diverse Fehler
			 */
			public void execute() throws Exception {
				// Siehe Feinkonzept
				flushCache();
				refreshIndex();
				
				_lastATime.accumulateAndGet(persMgr.maxATime(_containerFileDir), Math::max);
				_lastMediumId.accumulateAndGet(persMgr.maxMedienID(persMgr.getContainerHeaders(_indexLock, _containerFileDir.getArchiveDataKind())), Math::max);
				_maxContainerId.accumulateAndGet(persMgr.maxContainerId(_indexLock, _containerFileDir.getArchiveDataKind()), Math::max);
				
				// Zugriff sperren ist bereits durch Lock erfolgt
				copyFiles();
			}

			/**
			 * Schribt den Containerdateien-Schreibcache auf die Festplatte
			 * @throws Exception IO-Fehler
			 */
			private void flushCache() throws Exception {
				long openContID = persMgr.getOpenContID(_containerFileDir);
				if(openContID != -1) {
					CacheManager.Cache cache = CacheManager.getInstance().getCache(openContID);
					if(cache != null) {
						cache.flush();
					}
				}
			}

			/**
			 * Schreibt die Indexdateien im aktuellen Stand auf die Festplatte.
			 * @throws IndexException IO-Fehler
			 */
			private void refreshIndex() throws IndexException {
				persMgr.flushIndexes(_indexLock, _containerFileDir.getArchiveDataKind());
			}

			/**
			 * Kopiert die eigentlichen Dateien ins zielverzeichnis (ähnlich rsync)
			 * @throws IOException IO-Fehler
			 */
			private void copyFiles() throws IOException {
				Path path = persMgr.getPath(_containerFileDir);
				Path targetPath = _targetBaseDir.resolve(persMgr.getArchivePath().relativize(path));
				
				Files.createDirectories(targetPath);
				
				// Flag-Datei anlegen, damit Zielverzeichnis als Inkonsistent betrachtet wird, falls beim Kopieren das Archivsystem abstürzt o.ä.
				Path rebuildFlagFile = targetPath.resolve(PersistenceManager.REBUILDINDEX_FILE_FLAG_NAME);
				if(!Files.exists(rebuildFlagFile)) {
					Files.createFile(rebuildFlagFile);
				}

				SyncResult syncResult = new FileSync() {
					@Override
					protected boolean shouldIgnore(final String fileName) {
						switch(fileName) {
							case PersistenceManager.REBUILDINDEX_FILE_FLAG_NAME:
							case "_backupindex.idx":
								return true;
						}
						return fileName.endsWith(".defect");
					}
				}.syncDirectories(path, targetPath);
				
				_numFilesCopied.addAndGet(syncResult.getFilesCopied());
				_numFilesDeleted.addAndGet(syncResult.getFilesRemoved());
				_numBytesCopied.addAndGet(syncResult.getBytesCopied());

				// Flag-Datei wieder löschen falls alles bis hier geklappt hat
				Files.deleteIfExists(rebuildFlagFile);
			}
		}
	}

}
