/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import com.google.common.collect.Range;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.gap.GapFile;
import de.bsvrz.ars.ars.persistence.gap.RemoteArchive;
import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.ars.ars.persistence.iter.DeletedTreatment;
import de.bsvrz.ars.ars.persistence.iter.QueryDataSequence;
import de.bsvrz.ars.ars.persistence.walk.DataIdentificationDirWalk;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.util.Util;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Task zur Nachforderung fehlender Archivdaten aus anderen Archivsystemen gemäß Parametrierung.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RequestGapTask extends SingleClientSenderTask {
	
	/** Anzahl der Threads zum Nachfordern fehlender Archivdaten. */
	public static int REQUEST_GAP_THREADS = 1;

	/** Prioritaet der Nachforderungsanfragen. */
	private static final ArchiveQueryPriority REQUEST_PRIO = ArchiveQueryPriority.LOW;

	private final ArchiveDataKindCombination requestAdk = Util.getADKCombination(true, true, true, true);

	private ArchiveDataSpecification[] archDataSpecs;

	private List<RemoteArchive> archiveList;

	private final AtomicInteger totalGapsFound = new AtomicInteger(0);

	/**
	 * Zeitspezifikation, die beim Nachfordern verwendet werden soll oder {@code null}, falls keine Einschränkung berücksichtigt werden soll.
	 */
	private ArchiveTimeSpecification archTimeSpec;

	/**
	 * Erzeugt einen neuen Task zur Nachforderung fehlender Archivdaten.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange
	 */
	public RequestGapTask(ArchiveManager archiveMgr, BlockingQueue<Object> inputData) {
		super(archiveMgr, inputData);
	}

	/** 
	 * Liefert den Nachforderungsanfrage-Datensatz. 
	 * @return Dummy-Datensatz, der Nachforderung auslöst
	 */
	public static RequestGapResultData getRequestGapResultData() {
		return new RequestGapResultData();
	}

	/**
	 * Startet das automatische oder manuelle Nachfordern entsprechend dem übergebenen Nachforderungs-Datensatz.
	 *
	 * @param resultData Nachforderungs-Datensatz.
	 */
	@Override
	protected void work(ResultData resultData) {
		long t = Util.startTimer();
		totalGapsFound.set(0);
		long anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested();
		if(resultData instanceof RequestGapResultData) {
			_debug.info(getName() + ": Automatische Nachforderung gestartet.");
			archTimeSpec = null;
			archDataSpecs = new ArchiveDataSpecification[0];
			archiveList = new ArrayList<>(0);
			try {
				performRequest();
				anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested() - anzReqDS;
				_debug.info(
						getName() + ": Automatische Nachforderung beendet (Zeit: " + Util.stopTimer(t) + ", Luecken: " + totalGapsFound.get()
								+ ", erhaltene Datensaetze: " + anzReqDS + ")"
				);
			}
			catch(Exception e) {
				_debug.warning("Nachforderung fehlgeschlagen.", e);
			}
		}
		else {
			if(!resultData.hasData()) return;
			int msgType = resultData.getData().getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).intValue();
			if(msgType == MessageType.REQUEST_DID) {
				int messageTypeResult = MessageType.REQUEST_DID_RESULT;
				try {
					_debug.info(getName() + ": Manuelle Nachforderung gestartet.");
					init(resultData.getData());
					archTimeSpec = null;
					archDataSpecs = parseArchiveDataSpec(deserializer);
					archiveList = parseArchiveList(deserializer);
					performRequest();
					sendSuccessResponse(messageTypeResult);
					anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested() - anzReqDS;
					_debug.info(
							getName() + ": Manuelle Nachforderung beendet (Zeit: " + Util.stopTimer(t) + ", Luecken: " + totalGapsFound.get()
									+ ", erhaltene Datensaetze: " + anzReqDS + ")"
					);
				}
				catch(Exception e) {
					_debug.warning("Nachforderung fehlgeschlagen.", e);
					sendErrorResponse(messageTypeResult, "Nachforderung fehlgeschlagen: " + e.getMessage());
				}
			}
			else if(msgType == MessageType.REQUEST_AUTOM) {
				int messageTypeResult = MessageType.REQUEST_AUTOM_RESULT;
				try {
					_debug.info(getName() + ": Manueller Anstoß der automatischen Nachforderung.");
					init(resultData.getData());
					archDataSpecs = new ArchiveDataSpecification[0];
					final long startTime = deserializer.readLong();
					final long endTime = deserializer.readLong();
					archTimeSpec = new ArchiveTimeSpecification(TimingType.ARCHIVE_TIME, false, startTime, endTime);
					archiveList = parseArchiveList(deserializer);
					performRequest();
					sendSuccessResponse(messageTypeResult);
					anzReqDS = getArchMgr().getInQueuesMgr().getQueuesCountRequested() - anzReqDS;
					_debug.info(
							getName() + ": Manueller Anstoß der automatischen Nachforderung beendet (Zeit: " + Util.stopTimer(t) + ", Luecken: " + totalGapsFound
									.get()
									+ ", erhaltene Datensaetze: " + anzReqDS + ")"
					);
				}
				catch(Exception e) {
					_debug.warning("Nachforderung fehlgeschlagen.", e);
					sendErrorResponse(messageTypeResult, "Manueller Anstoß der automatischen Nachforderung fehlgeschlagen: " + e.getMessage());
				}
			}
			else {
				_debug.error("Ungültiger Nachrichtentyp für die Nachforderung: " + msgType);
			}
		}
		archDataSpecs = null;
	}

	/**
	 * Extrahiert die Liste der Archivsysteme, bei denen nachgefordert werden soll. Der Deserialisierer muss genau vor dem Längenbyte der Liste stehen.
	 * Doppelte EIntraege werden gefiltert, Nachfordern bei sich selbst ebenso.
	 *
	 * @param ds
	 * @return Liste mit Archivsystem-Objekten, leere Liste, falls keine angegeben sind
	 * @throws IOException Allgemeiner IO-Fehler
	 */
	private List<RemoteArchive> parseArchiveList(Deserializer ds) throws Exception {
		int anzArchives = deserializer.readInt(); // Größe der Liste mit den Archiven
		List<RemoteArchive> result = new ArrayList<>(anzArchives);

		HashSet<SystemObject> singleEntries = new HashSet<>();
		for(int i = 0; i < anzArchives; i++) {
			SystemObject foreignArS = ds.readObjectReference(archMgr.getDataModel());
			if(foreignArS == null) throw new IllegalArgumentException("Archivsystem-Objekt[" + i + "] == null");

			// doppelte Anfragen vermeiden, nicht bei sich selber nachfragen
			if(!singleEntries.contains(foreignArS) && !archMgr.getArchiveObject().getPid().equals(foreignArS.getPid())) {
				result.add(RemoteArchive.open(foreignArS));
				singleEntries.add(foreignArS);
			}
		}
		return result;
	}

	/**
	 * Erzeugt eine Anzahl ({@link #REQUEST_GAP_THREADS}) von RequestGapTask-Objekten, die die
	 * Nachforderung durchführen. Diese Methode wartet, bis alle Threads fertig und alle Daten archiviert sind.
	 */
	private void performRequest() throws PersistenceException {

		DataIdentificationDirWalk dirWalk;

		if(archDataSpecs.length == 0) {
			dirWalk = DataIdentificationDirWalk.parameterized(persMgr);
		}
		else {
			dirWalk = DataIdentificationDirWalk.create(persMgr, Arrays.asList(archDataSpecs));
		}

		dirWalk.execute("Indizierung für Nachfordern", REQUEST_GAP_THREADS, (dataIdentificationDir, walk) -> {
			if(shouldTerminate()) {
				walk.terminate();
				return;
			}
			
			Collection<ArchiveDataSpecification> ads = dataIdentificationDir.getArchiveDataSpecifications();
			
			if(ads != null) {
				for(ArchiveDataSpecification ad : ads) {
					if(shouldTerminate()) {
						walk.terminate();
						return;
					}
					requestDID(dataIdentificationDir.getDataIdentification(), dataIdentificationDir.getDataIdentNode(), ad.getTimeSpec());
				}
			}
			else {
				requestDID(dataIdentificationDir.getDataIdentification(), dataIdentificationDir.getDataIdentNode(), archTimeSpec);
			}
		});
	}

	/**
	 * Diese Klasse sieht aus wie ein ResultData, wird aber nur intern verwendet, um einen über den Scheduler angestoßenen Sicherungslauf von einem extern
	 * angestoßenen zu unterscheiden
	 *
	 * @author beck et al. projects GmbH
	 * @author Phil Schrettenbrunner
	 */
	public static class RequestGapResultData extends ResultData {

		/** Erzeugt ein neues leeres RequestGapResultData-Objekt. */
		public RequestGapResultData() {
			super(null, null, 0, null);
		}

	}

	/**
	 * Fordert die angegebene Datenidentifikation innerhalb des angegebenen Bereichs nach.
	 */
	private void requestDID(IdDataIdentification dataIdentification, @Nullable DataIdentNode din, @Nullable ArchiveTimeSpecification ats) throws PersistenceException {
		try {
			suspendTaskIfNecessary();    //Hält den Task an, falls das Archivsystem im kritischen Bereich.

			if(din == null) {
				din = archMgr.getDataIdentTree().getIfPresent(dataIdentification);
				if(din == null) {
					_debug.fine("DIN ist nicht parametriert für " + dataIdentification + ".");
					return;    // Überhaupt nicht parametriert
				}
			}

			if(!din.arSParamIsNachfordern()) {
				_debug.fine("Nachfordern ist nicht parametriert für " + dataIdentification + ".");
				return;    // Nachfordern nicht parametriert
			}

			// Nur für die parametrierten Archive nachfordern
			List<RemoteArchive> requestArchives;

			if(archiveList.size() > 0) {
				// Wenn eine ArchivListe übergeben wurde, die anzufragenden Archive darauf beschränken
				requestArchives = archiveList;
			}
			else {
				requestArchives = copyReqArSFromDIN(din);
			}

			// Alle Archive ignorieren, die nicht erreichbar sind
			requestArchives.removeIf(it -> {
				ArchiveRequestManager requestManager = it.createRequestManager(archMgr.getDavCon());
				return requestManager == null || !requestManager.isArchiveAvailable();
			});

			if(requestArchives.isEmpty()) {
				_debug.fine("Es ist kein Archiv verfügbar, um " + din.getDataIdentification() + " nachzufordern.");
				return;
			}

			AttributeGroup atg = archMgr.getAtg(dataIdentification.getAtgId());
			Aspect asp = archMgr.getAsp(dataIdentification.getAspectId());
			final SystemObject object = archMgr.getObj(dataIdentification.getObjectId());
			final DataDescription dataDesc = new DataDescription(atg, asp);

			DataIndexRangeSet gaps = new DataIndexRangeSet();


			try(SyncKey<IdDataIdentification> indexLock = archMgr.lockIndex(dataIdentification)) {

				QueryDataSequence dataSequence = new QueryDataSequence(archMgr, indexLock, ArchiveDataKindCombination
						.all(), ats, ArchiveOrder.BY_INDEX, DeletedTreatment.Full);

				try(DataIterator iterator = dataSequence.iteratorWithoutTimeFilter(null)) {

					if(iterator.isEmpty()) {
						_debug.fine("Es sind keine bestehenden Daten archiviert, " + din.getDataIdentification() + " kann daher nicht nachgefordert werden.");
						return;
					}

					TimingType timingType = ats == null ? TimingType.DATA_INDEX : ats.getTimingType();

					long lowDidx = -1;
					long lowFilterValue = -1;

					while(!iterator.isEmpty()) {

						if(shouldTerminate()) return;

						long hiDidx = iterator.peekDataIndex();
						long highFilterValue = peekFilterValue(timingType, iterator);
						
						if(isGap(lowDidx, hiDidx)) {
							// Es gibt eine Lücke
							
							if(isInRange(ats, lowFilterValue, highFilterValue)) {
								// Die Lücke liegt im Anfragebereich
								
								totalGapsFound.incrementAndGet();

								gaps.addRange(noModBits(lowDidx) * 4 + 4, hiDidx);
							}

						}
						lowDidx = hiDidx;
						lowFilterValue = highFilterValue;
						iterator.remove();
					}
				}
			}
			if(gaps.isEmpty()) {
				_debug.fine("Keine Datenlücken vorhanden für '" + dataIdentification + "'");
				return;
			}
			requestGapData(gaps, requestArchives, dataIdentification, dataDesc, object);
		}
		catch(Exception e) {
			_debug.warning("Fehler beim Nachfordern von '" + dataIdentification, e);
		}

		try {
			_debug.fine("Warte auf Fertigstellung der Nachforderung.");
			Semaphore semaphore = new Semaphore(1);
			semaphore.acquire();
			archMgr.getInQueuesMgr().insertInArchiveQueueReq(System.currentTimeMillis(), semaphore);
			semaphore.acquire();
			_debug.fine("Nachforderung fertig.");
		}
		catch(InterruptedException e) {
			throw new PersistenceException("Warten auf Nachfordern wurde unterbrochen");
		}
	}

	private static boolean isInRange(@Nullable final ArchiveTimeSpecification ats, final long a, final long b) {
		if(ats == null) return true;
		long start = ats.getIntervalStart();
		long end = ats.getIntervalEnd();
		if(a <= b) {
			return a <= end && b >= start;
		}
		return b <= end && a >= start;
	}

	private static long peekFilterValue(final TimingType timingType, final DataIterator iterator) {
		if(timingType == TimingType.DATA_TIME) {
			return iterator.peekDataTime();
		}
		if(timingType == TimingType.ARCHIVE_TIME) {
			return iterator.peekArchiveTime();
		}
		if(timingType == TimingType.DATA_INDEX) {
			return iterator.peekDataIndex();
		}
		throw new IllegalArgumentException("TimingType: " + timingType);
	}

	private void requestGapData(final DataIndexRangeSet gaps, final List<RemoteArchive> archivesToRequestFrom, final IdDataIdentification dataIdentification, final DataDescription dataDesc, final SystemObject object) {
		_debug.fine("Starte Nachforderung von " + gaps.size() + " Lücken bei " + archivesToRequestFrom);
		DataIndexRangeSet remainingGaps = new DataIndexRangeSet(gaps);
		DataIndexRangeSet allClosedGaps = new DataIndexRangeSet();

		GapFile gapFile = GapFile.getInstance(persMgr, dataIdentification);
		
		final Map<RemoteArchive, DataIndexRangeSet> previouslyFailedGapsPerArs = new HashMap<>();
		
		gapFile.readGaps(previouslyFailedGapsPerArs);
		
		final Map<RemoteArchive, DataIndexRangeSet> failedGapsPerArs = new HashMap<>(previouslyFailedGapsPerArs);
		
		for(RemoteArchive archiveObject : archivesToRequestFrom) {
			DataIndexRangeSet previouslyFailedGaps = previouslyFailedGapsPerArs.get(archiveObject);
			if(previouslyFailedGaps == null) {
				previouslyFailedGaps = new DataIndexRangeSet();
			}
			DataIndexRangeSet gapsToRequest = new DataIndexRangeSet(remainingGaps);
			// Keine Daten anfragen, wo wir schon wissen, dass das Archivsystem diese nicht besitzt.
			// Ausnahme: Wenn Archivsysteme explizit angegeben sind, wird dort auch dann nachgefragt, wenn dies
            // bereits als nicht erfolgreich vermerkt wurde (TAnfArS 5.1.2.12.2)
			if(archiveList.isEmpty()) {
				gapsToRequest.removeAll(previouslyFailedGaps);
			}
			DataIndexRangeSet closedGaps = requestGapData(gapsToRequest, archiveObject, dataDesc, object);
			allClosedGaps.addAll(closedGaps);
			DataIndexRangeSet failedGaps = new DataIndexRangeSet(gapsToRequest);
			failedGaps.removeAll(closedGaps);
			failedGaps.addAll(previouslyFailedGaps);
			failedGapsPerArs.put(archiveObject, failedGaps);
			remainingGaps.removeAll(closedGaps);
			if(remainingGaps.isEmpty()) {
				_debug.fine("Alle Lücken sind beim Nachfordern geschlossen worden, weitere Archivsysteme werden ignoriert.");
				break;
			}
		}

		// Datenlücken die geschlossen wurden, brauchen nicht mehr im Gapfile gespeichert zu werden, da wir uns nicht mehr dafür interessieren
		// (Sie würden eh nicht neu angefragt, weil die Daten ja jetzt da sind)
		for(DataIndexRangeSet value : failedGapsPerArs.values()) {
			value.removeAll(allClosedGaps);
		}
		
		gapFile.writeGaps(failedGapsPerArs);
		
	}

	/**
	 * Fordert Daten von einem Archivsystem nach
	 * @param gaps          Liste mit Datenlücken
	 * @param archiveSystem Archivsystem, von dem nachgefordert werden soll
	 * @param dataDesc      {@link DataDescription}, die nachgefordert werden soll
	 * @param object        Objekt, für das nachgefordert werden soll
	 * @return Alle Datenindexbereiche der Datenlücken, die mit diesem Archivsystem geschlossen werden konnten
	 */
	private DataIndexRangeSet requestGapData(final DataIndexRangeSet gaps, final RemoteArchive archiveSystem, final DataDescription dataDesc, final SystemObject object) {

		DataIndexRangeSet closedGaps = new DataIndexRangeSet();
		if(gaps.isEmpty()) {
			return closedGaps; // Nichts zu tun, keine Lücken da
		}
		
		ArchiveRequestManager arm = archiveSystem.createRequestManager(archMgr.getDavCon());
		
		if(arm == null) {
			_debug.fine("Nachfordern: Archivsystem '" + archiveSystem + "' ist nicht erreichbar");
			// Archivsystem nicht erreichbar
			return closedGaps;
		}

		for(Range<Long> gap : gaps) {

			Long lowBnd = gap.lowerEndpoint();
			Long upBnd = gap.upperEndpoint();

			try {
				ArchiveDataQueryResult res = arm.request(REQUEST_PRIO, buildRequest(lowBnd, upBnd, dataDesc, object));
				if(!res.isRequestSuccessful()) {
					_debug.warning("Nachforderungsanfrage an Archivsystem '" + archiveSystem + "' fehlgeschlagen", res.getErrorMessage());
					return closedGaps;
				}
				
				ArchiveDataStream ads = res.getStreams()[0];    // nur einer war angelegt
				try {
					ArchiveData ad;
					while((ad = ads.take()) != null) {
						if(idx(ad) >= upBnd) break;         // drüber (letzter)

						if(idx(ad) < lowBnd) {
							continue;    // drunter, erster oder potGap mit gleichem Index wie letzter DS
						}

						// Datenindex passt. Nur speichern, wenn korrekte Datenart. Lückenmarkierungen
						// die einen fehlerhaften Index haben, werden dadurch nicht gespeichert.
						if(properDataState(ad)){
							archMgr.getInQueuesMgr().insertInArchiveQueueReq(System.currentTimeMillis(), ad);
							closedGaps.addRange(ad.getDataIndex(), noModBits(ad.getDataIndex()) * 4 + 4);
						}
					}
				}
				catch(IOException e) {
					_debug.warning("Fehler beim Empfang eines Nachforderungs-Datensatzes von '" + archiveSystem + "'", e);
				}
				finally {
					ads.abort();
				}
			}
			catch(IllegalStateException | InterruptedException e) {
				_debug.fine("Nachfordern: Archivsystem '" + archiveSystem + "' ist nicht erreichbar", e);
			}
		}
		return closedGaps;
	}


	/**
	 * Kopiert die Liste der Nachforderungs-Archive aus dem DataIdentNode. Jedes Archivsystem wird nur einmal in die Ergebnisliste aufgenommen.
	 *
	 * @param din
	 * @return Liste von Archiv-Objekten
	 */
	private List<RemoteArchive> copyReqArSFromDIN(DataIdentNode din) {
		ArrayList<RemoteArchive> otherArchives = new ArrayList<>();
		HashSet<String> singleEntries = new HashSet<>();
		for(int i = 0; i < din.arSParamGetAnzNachfordern(); i++) {
			String pid = din.arSParamGetNachfordern(i);

			if(pid == null || pid.isEmpty() || pid.equals("0")) continue;    // DAV erlaubt hier Nullwerte

			// doppelte Anfragen vermeiden, nicht bei sich selber nachfragen
			if(!singleEntries.contains(pid) && !archMgr.getConfigAuth().getPid().equals(pid)) {
				otherArchives.add(RemoteArchive.open(pid));
				singleEntries.add(pid);
			}
		}
		return otherArchives;
	}

	// ***************************************************************************************************
	// **********  Abkuerzende Methoden (um den Hauptalgorithmus einfach zu halten) **********************
	// ***************************************************************************************************

	private ArchiveDataSpecification buildRequest(long startIdx, long endIdx, final DataDescription dataDesc, final SystemObject object) {
		return new ArchiveDataSpecification(
				new ArchiveTimeSpecification(TimingType.DATA_INDEX, false, startIdx, endIdx),
				requestAdk,
				ArchiveOrder.BY_INDEX,
				ArchiveRequestOption.NORMAL,
				dataDesc,
				object
		);
	}

	private static boolean isGap(long start, long end) {
		return start != -1 && noModBits(end) - noModBits(start) > 1;
	}

	private static boolean properDataState(ArchiveData ad) {
		DataState ds = ad.getDataType();
		return ds.equals(DataState.DATA) || ds.equals(DataState.NO_DATA) || ds.equals(DataState.NO_SOURCE);
	}

	private static long idx(ArchiveData ad) {
		return ad.getDataIndex();
	}

	private static long noModBits(long dataIndex) {
		return Util.dIdxNoModBits(dataIndex);
	}
}
