/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import static java.nio.charset.StandardCharsets.ISO_8859_1;


import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerFileHandle;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.ContainerHeaders;
import de.bsvrz.ars.ars.persistence.ContainerManagementInformation;
import de.bsvrz.ars.ars.persistence.DataIdentificationDir;
import de.bsvrz.ars.ars.persistence.DeletedContainerFile;
import de.bsvrz.ars.ars.persistence.IdContainerFileDir;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.IndexValues;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.ars.ars.persistence.walk.ContainerAction;
import de.bsvrz.ars.ars.persistence.walk.ContainerWalk;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.BlockingQueue;

/**
 * Task zur Wiederherstellung gesicherter Container.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RestoreTask extends SingleClientSenderTask {

	private static final String UNAVAIL_MEDIUM_ID_FILE = "_unavailable_mediumids.dat";

	// parametrierte Löschschutzverlängerung in Sekunden
	private static long LOESSCHUTZ_VERL;

	private final Set<String> mediumIdIgnoreSet = new HashSet<>();

	/**
	 * Objekt, dem Informationen zum aktuellen Zustand der Task und dem Fortschritt von langwierigen Arbeiten übergeben werden.
	 */
	private final LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;

	/**
	 * Erzeugt einen neuen Task zur Wiederherstellung gesicherter Container.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange
	 * @param longTermTaskStatePublisher  Klasse, die Taskzustände veröffentlicht
	 */
	public RestoreTask(ArchiveManager archiveMgr, BlockingQueue<Object> inputData, final LongTermTaskStatePublisher longTermTaskStatePublisher) {
		super(archiveMgr, inputData);
		setName("Restore Task");
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Wiederherstellung");
	}


	@Override
	public void execute() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.execute();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	/**
	 * Prüft, ob der angegebene Dateiname der Name einer _unavailable_mediumids.dat-Datei ist.
	 *
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer _unavailable_mediumids.dat-Datei ist
	 */
	public static boolean isUnavailMediumIDsFile(String name) {
		return name.equals(UNAVAIL_MEDIUM_ID_FILE);
	}

	/**
	 * Fuehrt die Wiederherstellung gesicherter Datenvontainer durch.
	 *
	 * @param resultData Zu verarbeitender Datensatz aus der Eingangswarteschlange.
	 */
	@Override
	protected void work(ResultData resultData) {
		if(!resultData.hasData()) return;
		try {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);
			init(resultData.getData());
			initIgnoreList();
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
			restore();
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
			sendSuccessResponse();
		}
		catch(MediumNotAccessibleException e) {
			_debug.warning("Wiederherstellung abgebrochen: Medium nicht verfügbar.", e);
			MessageSender.getInstance().sendMessage(
					"", MessageType.APPLICATION_DOMAIN, "", MessageGrade.WARNING, MessageState.MESSAGE, "Bitte legen Sie Medium ein"
			);
		}
		catch(Exception e) {
			_debug.error("Wiederherstellung fehlgeschlagen.", e);
			sendErrorResponse("Wiederherstellung fehlgeschlagen: " + e.getMessage());
		}
		finally {
			mediumIdIgnoreSet.clear();
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}

	private void restore() throws Exception {

		// Wir bauen eine Liste mit Medien-IDs und Container-Filenamen auf.
		// Diese Liste sortieren wir danach nach Medien-ID und stellen jeden einzelnen Container wieder her.
		List<MediumidContainerTupel> medienIds;

		long numRestored = 0;

		ArchiveDataSpecification[] ads = parseArchiveDataSpec(deserializer);
		if(ads.length == 0) {
			medienIds = collectAllMediumIdsFromDir();    // Methode synchronisiert selbst auf DataIdentNode
		}
		else {
			medienIds = collectMediumIdsByDataSpec(ads); // Methode synchronisiert selbst auf DataIdentNode
		}

		Collections.sort(medienIds);
		if(shouldTerminate()) throw new IllegalStateException("Archivsystem wurde beendet. Wiederherstellungslauf abgebrochen");

		long protectionTime = System.currentTimeMillis() + LOESSCHUTZ_VERL * 1000;

		// Container wiederherstellen:
		for(MediumidContainerTupel tupel : medienIds) {
			suspendTaskIfNecessary();	//Task anhalten, falls das Archivsystem überlastet.

			Path contFile = tupel.getContainerFile();
			int mediumID = tupel.getMediumID();
			// Synchronisierung über DataIdentNode:
			Path parent = contFile.getParent();
			IdDataIdentification did = PersistenceManager.getDataIdentificationFromPath(parent);
			ArchiveDataKind adk = PersistenceManager.getArchiveDataKindFromDir(parent);

			restoreFromBackup(mediumID, did, adk, contFile.toAbsolutePath().toString(), protectionTime);

			_longTermTaskStatePublisherTask.setTaskProgress("Wiederherstellung: Von %1$d von %2$d Medien wiederhergestellt, letzte MedienID: %3$d", ++numRestored, medienIds
					.size(), mediumID);

			if(shouldTerminate()) throw new IllegalStateException("Archivsystem wurde beendet. Wiederherstellungslauf abgebrochen");
		}
		_debug.info("Wiederherstellung erfolgreich beendet: " + medienIds.size() + " Container wiederhergestellt.");
	}

	private List<MediumidContainerTupel> collectAllMediumIdsFromDir() throws Exception {
		List<MediumidContainerTupel> medienIds = new ArrayList<>();
		// Über alle Verzeichnisse iterieren:

		ContainerWalk.allContainers(persMgr).execute("Wiederherstellung", 1, new ContainerAction() {
			long directoriesVisitedCount;

			@Override
			public void run(final DataIdentificationDir dataIdentificationDir, final SyncKey<IdDataIdentification> lock, final ArchiveDataKind adk, final Collection<ContainerHeaders> headers) {
				for(ContainerHeaders header : headers) {
					if(shouldTerminate()) throw new IllegalStateException("Archivsystem wurde beendet. Wiederherstellunglauf abgebrochen");
					final String mediumID = header.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID);
					if(isMediumIdToBeRestored(mediumID)) {
						long contId = header.getContainerHeaderParamAsLong(ContainerManagementInformation.CHP_CONT_ID);
						medienIds.add(new MediumidContainerTupel(Integer.parseInt(mediumID), dataIdentificationDir.getDataIdentificationDir().resolve(PersistenceManager.getDataKindSuffix(adk)).resolve(PersistenceManager.getContFileName(contId))));
					}
				}
				_longTermTaskStatePublisherTask.setTaskProgress("Indizierung: %1$d Verzeichnisse untersucht, %3$d wiederherzustellende Container gefunden", ++directoriesVisitedCount, -1, medienIds
						.size());
			}
		});
		return medienIds;
	}

	private List<MediumidContainerTupel> collectMediumIdsByDataSpec(ArchiveDataSpecification[] ads) throws Exception {
		List<MediumidContainerTupel> medienIds = new ArrayList<>();

		for(int i = 0; i < ads.length; i++) {
			ArchiveDataSpecification arcDataSpec = ads[i];

			ArchiveTimeSpecification atSpec = arcDataSpec.getTimeSpec();

			// Prüfen, ob jemals Daten für diese Datenidentifikation archiviert wurden:
			for(ArchiveDataKind dataKind : arcDataSpec.getDataKinds()) {

				try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(new IdDataIdentification(arcDataSpec))) {

					// Synchronisiserter Index-Zugriff:
					IndexResult<IndexValues> ir = null;
					if(persMgr.numContainers(lock, dataKind) > 0) {
						ir = persMgr.getIndexResult(lock, dataKind, atSpec);
					}

					if(ir != null) {
						// Alle Container inklusive der Medien-ID sammeln:
						for(int j = 0; j < ir.size(); j++) {
							long containerId = ir.get(j, IndexValues.ContainerId);
							Path contFile =
									persMgr.getPath(arcDataSpec, dataKind).resolve(PersistenceManager.getContFileName(containerId));
							ContainerHeaders headers = persMgr.getContainerHeaders(lock, dataKind, containerId);
							boolean deleted = headers.getContainerHeaderParamAsBoolean(ContainerManagementInformation.CHP_DELETED);
							final String mediumID = headers.getContainerHeaderParamAsString(ContainerManagementInformation.CHP_MEDIUM_ID);
							if(deleted && isMediumIdToBeRestored(mediumID)) {
								medienIds.add(new MediumidContainerTupel(Integer.parseInt(mediumID), contFile));
							}
							if(shouldTerminate()) throw new IllegalStateException("Archivsystem wurde beendet. Wiederherstellunglauf abgebrochen");
						}
					}
				}

			}
			if(shouldTerminate()) throw new IllegalStateException("Archivsystem wurde beendet. Wiederherstellunglauf abgebrochen");
			_longTermTaskStatePublisherTask.setTaskProgress("Indizierung: %1$d von %2$d Datenspezifikationen untersucht, %3$d wiederherzustellende Container gefunden", i + 1, ads.length, medienIds.size());
		}
		return medienIds;
	}

	/** Holt einen Container aus dem Backup zurück. */
	private void restoreFromBackup(int backupMediumID, final IdDataIdentification dataIdentification, ArchiveDataKind adk, String containerFileName, final long protectionTime)
			throws Exception {

		// Zeit messen:
		long startTime = Util.startTimer();

		containerFileName = containerFileName.substring(containerFileName.lastIndexOf(File.separator) + 1);
	
		try(SyncKey<IdDataIdentification> lock = archMgr.lockIndex(dataIdentification)) {

			// Die Containerdatei muss mit Lock geschrieben werden (falls z. B. eine Archivanfrage parallel auf die Daten läuft)
			// InputStream oeffnen:
			try (InputStream is = BackupTask.getBackupModul().restoreContainer(backupMediumID, containerFileName)) {
				if (is == null) {
					_debug.warning("Container " + containerFileName + " nicht gefunden.");
					return;
				}

				Path file = persMgr.getPath(dataIdentification, adk).resolve(containerFileName);
				try (OutputStream fos = Files.newOutputStream(file)) {
					Util.copyStreams(is, fos);
				}
			} catch (BackupException e) {
				MessageSender.getInstance().sendMessage("", MessageType.APPLICATION_DOMAIN, "", MessageGrade.ERROR, MessageState.MESSAGE,
				                                        "Wiederherstellung fehlgeschlagen");
				_debug.error("Wiederherstellung fehlgeschlagen.", e);
				throw e;
			}

			try (ContainerFileHandle handle = persMgr
				.accessContainer(new IdContainerFileDir(lock.getElement(), adk), ContainerFile.getContID(containerFileName))) {

				// Containerheader anpassen auf "Wiederhergestellt"

				handle.readContainerHeader();
				handle.setContainerHeaderParam(ContainerHdr.CHP_LOESCHUTZ, protectionTime);
				handle.setContainerHeaderParam(ContainerHdr.CHP_RESTORED, "1");
				handle.setContainerHeaderParam(ContainerHdr.CHP_MEDIUM_ID, Util.leadZero(backupMediumID, ContainerHdr.CHP_MEDIUM_ID.getValLen()));

				handle.writeContainerHeader(lock);

				DeletedContainerFile deletedContainerFile = persMgr.openDeletedContainerFile(lock, adk);
				deletedContainerFile.removeDeletedContainer(handle.getContainerId());
				deletedContainerFile.write();
			}

			// Zeit messen & Log-Ausgabe:
			_debug.info("Container " + containerFileName + " erfolgreich wiederhergestellt (" + Util.stopTimer(startTime) + " sec.)");
		}
	}

	/**
	 * Liest die Datei MEDIUM_ID_IGNORE_FILE aus und schreibt die darin enthaltenen MedienIds in das Set mediumIdIgnoreSet. Fehler werden großzügig behandelt -
	 * notfalls bleibt das Set leer.
	 */
	private void initIgnoreList() {
		Path ignoreFile = persMgr.getArchivePath().resolve(UNAVAIL_MEDIUM_ID_FILE);
		if(!Files.exists(ignoreFile)) {
			_debug.info("Datei " + UNAVAIL_MEDIUM_ID_FILE + " existiert nicht.");
			return;
		}
		try(BufferedReader br = Files.newBufferedReader(ignoreFile, ISO_8859_1)) {
			while(true) {
				String s = br.readLine();
				if(s != null) {
					try {
						int mediumID = Integer.parseInt(s.trim());
						mediumIdIgnoreSet.add(String.valueOf(mediumID));
					}
					catch(NumberFormatException ignored) {
						// do nothing
					}
				}
				else {
					break;	//Dateiende erreicht
				}
				suspendTaskIfNecessary();	//Task ggf. anhalten
			}
		}
		catch(IOException e) {
			_debug.warning(
					"Exception trat beim Lesen der Datei " + UNAVAIL_MEDIUM_ID_FILE + " auf. Wiederherstellung verwendet keine MedienID-Ignore-Liste.", e
			);
		}
	}

	private boolean isMediumIdToBeRestored(String mediumID) {
		return !mediumID.equals(ContainerHdr.MEDIUM_DESTROYED) && !mediumID.equals(ContainerHdr.MEDIUM_UNSAVED) && !mediumIdIgnoreSet.contains(mediumID);
	}

	/** 
	 * Gibt parametrierte Löschschutzverlängerung in Sekunden zurück
	 * @return parametrierte Löschschutzverlängerung in Sekunden
	 */
	public static long getLOESSCHUTZ_VERL() {
		return LOESSCHUTZ_VERL;
	}

	/**
	 * Setzt die parametrierte Löschschutzverlängerung in Sekunden
	 * @param loesschutz_verl Löschschutzverlängerung
	 */
	public static void setLOESSCHUTZ_VERL(long loesschutz_verl) {
		LOESSCHUTZ_VERL = loesschutz_verl;
	}

	private static class MediumidContainerTupel implements Comparable<MediumidContainerTupel> {

		private final int mediumID;

		private final Path containerFile;

		public MediumidContainerTupel(int mediumid, Path file) {
			containerFile = file;
			mediumID = mediumid;
		}

		public Path getContainerFile() {
			return containerFile;
		}

		public int getMediumID() {
			return mediumID;
		}

		@Override
		public int compareTo(@NotNull MediumidContainerTupel o) {
			int otherMediumID = o.getMediumID();
			return Integer.compare(mediumID, otherMediumID);
		}

		@Override
		public boolean equals(final Object o) {
			if(this == o) return true;
			if(o == null || getClass() != o.getClass()) return false;

			final MediumidContainerTupel that = (MediumidContainerTupel) o;

			if(mediumID != that.mediumID) return false;
			return Objects.equals(containerFile, that.containerFile);
		}

		@Override
		public int hashCode() {
			int result = mediumID;
			result = 31 * result + (containerFile != null ? containerFile.hashCode() : 0);
			return result;
		}
	}

}
