/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks.scheduler;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.LZVTask;
import de.bsvrz.ars.ars.mgmt.tasks.RequestGapTask;
import de.bsvrz.sys.funclib.debug.Debug;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Diese Klasse verwaltet einen Timer und prüft im Minutentakt, ob die Tasks für Sicherung, Löschen und Nachforndern ausgeführt werden müssen.
 * <p>
 * Über die public-Methoden können die IntervalWerte gesetzt werden.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class TaskScheduler {

	private static final ThreadLocal<DateFormat> dateFormat = ThreadLocal.withInitial(() -> new SimpleDateFormat("dd.MM.yyyy HH:mm:ss"));

	private static final int ONE_MINUTE = 60 * 1000;
	
	private static final long ONE_HOUR = 60 * ONE_MINUTE;

	private static Debug logger;

	private boolean terminated;

	private final Timer timer;

	private long capacityCheckInterval;

	private final ArchiveManager archiveManager;

	private final TimeIntervalCron backupTimeInterval = new TimeIntervalCron();

	private final TimeIntervalCron deleteTimeInterval = new TimeIntervalCron();

	private final TimeIntervalCron deletePermanentlyTimeInterval = new TimeIntervalCron();

	private final TimeIntervalCron requestTimeInterval = new TimeIntervalCron();

	private Date backupLastRun;

	private Date deleteLastRun;

	private Date deletePermanentlyLastRun;

	private Date requestLastRun;

	private Date lzvLastRun;

	private final CronTask backupCronTask;

	private final CronTask deleteCronTask;

	private final CronTask deletePermanentlyCronTask;

	private final CronTask requestCronTask;

	private TimerTask capacityCheckTimerTask;

	private final TimerTask saveUnsubscriptionTimeTask;


	/**
	 * Erzeugt einen TaskScheduler mit den angegebenen Interval-Zeiten für den Sicherungs-, Lösch- und Nachforderungstask.
	 *
	 * @param archiveManager Archiv-Manager
	 */
	public TaskScheduler(ArchiveManager archiveManager) {
		logger = Debug.getLogger();
		this.timer = new Timer();
		this.archiveManager = archiveManager;
		backupCronTask = new CronTask(backupTimeInterval);
		deleteCronTask = new CronTask(deleteTimeInterval);
		deletePermanentlyCronTask = new CronTask(deletePermanentlyTimeInterval);
		requestCronTask = new CronTask(requestTimeInterval);
		saveUnsubscriptionTimeTask = new UnsubscriptionTimerTask();
	}


	/** Startet den Timer, der die Tasks steuert */
	public void start() {
		// Timer zur nächsten vollen Minute starten
		long start = ONE_MINUTE - (System.currentTimeMillis() % ONE_MINUTE);
		start = start < 100 ? 0 : start;
		timer.scheduleAtFixedRate(backupCronTask, start, ONE_MINUTE);
		timer.scheduleAtFixedRate(deleteCronTask, start, ONE_MINUTE);
		timer.scheduleAtFixedRate(deletePermanentlyCronTask, start, ONE_MINUTE);
		timer.scheduleAtFixedRate(requestCronTask, start, ONE_MINUTE);
		timer.scheduleAtFixedRate(saveUnsubscriptionTimeTask, start, ONE_HOUR);
	}


	/**
	 * Setzt die Intervalle
	 *
	 * @param backup                Backup-Intervall (Cron-String)
	 * @param delete                Lösch-Intervall (Cron-String)
	 * @param request               Nachforderungs-Intervall (Cron-String)
	 * @param capacityCheckInterval Intervall für Kapazitätsprüfung in ms
	 */
	public void setIntervals(String backup, String delete, String request, long capacityCheckInterval) {
		backupTimeInterval.setFields(backup);
		if(backupTimeInterval.hasParseErrors()) {
			logger.warning(
					"'" + backup + "' ist kein gültiger ZeitIntervallCron für das Sicherungsintervall. Automatischer Start für die Sicherung DEAKTIVIERT"
			);
		}
		else {
			logger.fine("Setze Sicherungsintervall auf: " + backupTimeInterval.toString());
		}

		deleteTimeInterval.setFields(delete);
		if(deleteTimeInterval.hasParseErrors()) {
			logger.warning("'" + delete + "' ist kein gültiger ZeitIntervallCron für das Löschintervall. Automatischer Start für das Löschen DEAKTIVIERT");
		}
		else {
			logger.fine("Setze Löschintervall auf: " + deleteTimeInterval.toString());
		}

		requestTimeInterval.setFields(request);
		if(requestTimeInterval.hasParseErrors()) {
			logger.warning(
					"'" + request + "' ist kein gültiger ZeitIntervallCron für das Nachforderungsintervall. Automatischer Start für das Nachfordern DEAKTIVIERT"
			);
		}
		else {
			logger.fine("Setze Nachforderungsintervall auf: " + requestTimeInterval.toString());
		}

		if(capacityCheckTimerTask != null) capacityCheckTimerTask.cancel(); // beendet ev. vorhandenen Timer
		if(capacityCheckInterval < ONE_MINUTE) capacityCheckInterval = ONE_MINUTE; // 1 Min ist kleinster Wert
		capacityCheckTimerTask = new CapacityCheckTimerTask();
		if(!terminated) timer.scheduleAtFixedRate(capacityCheckTimerTask, capacityCheckInterval, capacityCheckInterval);
		this.capacityCheckInterval = capacityCheckInterval;
	}


	/**
	 * Setzt das Interval fürs endgültige Löschen
	 * @param permanentlyDeleteCron Cron-String
	 * @param enabled               Aktiviert?
	 */
	public void setDeletePermanentlyInterval(String permanentlyDeleteCron, boolean enabled) {
		if(!enabled) {
			deletePermanentlyTimeInterval.setFields("");
		}
		else {
			deletePermanentlyTimeInterval.setFields(permanentlyDeleteCron);
		}
		if(deletePermanentlyTimeInterval.hasParseErrors()) {
			logger.warning("'" + permanentlyDeleteCron + "' ist kein gültiger ZeitIntervallCron für das endgültige Löschintervall. Automatischer Start für das endgültige Löschen DEAKTIVIERT");
		}
		else {
			logger.fine("Setze Intervall für das endgültige Löschen auf: " + deleteTimeInterval.toString());
		}
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgeführt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getBackupIntervals() {
		return backupTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getBackupLastRun() {
		if(backupLastRun != null) return dateFormat.get().format(backupLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nächste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungültig sind). Berücksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rückgabewert "&gt; 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getBackupNextRun() {
		if(backupTimeInterval.hasValues()) {
			Calendar backupNextRun = backupTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(backupNextRun != null) {
				return dateFormat.get().format(backupNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgeführt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getDeleteIntervals() {
		return deleteTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getDeleteLastRun() {
		if(deleteLastRun != null) return dateFormat.get().format(deleteLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nächste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungültig sind). Berücksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rückgabewert "&gt; 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getDeleteNextRun() {
		if(deleteTimeInterval.hasValues()) {
			Calendar deleteNextRun = deleteTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(deleteNextRun != null) {
				return dateFormat.get().format(deleteNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgeführt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getDeletePermanentlyIntervals() {
		return deletePermanentlyTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getDeletePermanentlyLastRun() {
		if(deletePermanentlyLastRun != null) return dateFormat.get().format(deletePermanentlyLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nächste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungültig sind). Berücksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rückgabewert "&gt; 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getDeletePermanentlyNextRun() {
		if(deletePermanentlyTimeInterval.hasValues()) {
			Calendar deletePermanentlyNextRun = deletePermanentlyTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(deletePermanentlyNextRun != null) {
				return dateFormat.get().format(deletePermanentlyNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgeführt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getRequestIntervals() {
		return requestTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getRequestLastRun() {
		if(requestLastRun != null) return dateFormat.get().format(requestLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nächste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungültig sind). Berücksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rückgabewert "&gt; 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getRequestNextRun() {
		if(requestTimeInterval.hasValues()) {
			Calendar requestNextRun = requestTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(requestNextRun != null) {
				return dateFormat.get().format(requestNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval (ausgewertet) als String. Wenn leerer String "", wird der Task nie ausgeführt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getLZVIntervals() {
		if(capacityCheckInterval == 0) {
			return "";
		}
		else if(capacityCheckInterval < ONE_HOUR) {
			return (capacityCheckInterval / ONE_MINUTE) + " Minuten";
		}
		else {
			return (capacityCheckInterval / ONE_HOUR) + " Stunden";
		}
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getLZVLastRun() {
		if(lzvLastRun != null) return dateFormat.get().format(lzvLastRun);
		return "(wurde noch nie gestartet)";
	}


	/** TaskScheduler beenden. Laufende Tasks laufen aber noch zu Ende. Es werden keine neuen Tasks mehr gestartet. */
	public void terminate() {
		terminated = true;
		if(backupCronTask != null) backupCronTask.cancel();
		if(deleteCronTask != null) deleteCronTask.cancel();
		if(deletePermanentlyCronTask != null) deletePermanentlyCronTask.cancel();
		if(requestCronTask != null) requestCronTask.cancel();
		if(capacityCheckTimerTask != null) capacityCheckTimerTask.cancel();
		if(saveUnsubscriptionTimeTask != null) saveUnsubscriptionTimeTask.cancel();
		timer.cancel();
	}


	/**
	 * Private Klasse, die sich um den Aufruf des spontanen Loeschen kümmert. Da dieser Task periodisch gestartet wird, ist er in einer eigenen Klasse
	 * implementiert
	 *
	 * @author Philippe Schrettenbrunner
	 */
	private class CapacityCheckTimerTask extends TimerTask {

		@Override
		public void run() {
			logger.info("Starte Kapazitätscheck");
			try {
				archiveManager.getInQueuesMgr().insertInLZVQueue(LZVTask.getCapacityCheckResultData());
				lzvLastRun = new Date(scheduledExecutionTime());
			}
			catch(RuntimeException e) {
				logger.warning("Fehler beim Start des Kapazitätscheck. Der Kapazitätscheck (und spontanes Löschen) wird nicht ausgeführt", e);
			}
		}
	}

	/**
	 * Private Klasse, die sich um den Aufruf des spontanen Loeschen kümmert. Da dieser Task periodisch gestartet wird, ist er in einer eigenen Klasse
	 * implementiert
	 *
	 * @author Philippe Schrettenbrunner
	 */
	private class UnsubscriptionTimerTask extends TimerTask {

		@Override
		public void run() {
			archiveManager.getPersMgr().saveUnsubscriptionTime();
		}
	}


	/**
	 * Private Klasse, die sich um den Aufruf von Sichern, Löschen und Nachfordern kümmert. Die Klasse wird minütlich aufgerufen und in der run()-Methode wird
	 * gepfüft, ob einer der Tasks gestartet werden muss (auf Grund der Intervallangabe im TimeIntervalCron)
	 *
	 * @author Philippe Schrettenbrunner
	 */
	private class CronTask extends TimerTask {

		private final TimeIntervalCron tic;

		public CronTask(TimeIntervalCron tic) {
			this.tic = tic;
		}


		@Override
		public void run() {
			Date d = new Date(scheduledExecutionTime());
			if(tic.shouldRun(d)) {
				if(tic == backupTimeInterval) {
					logger.info("Starte Sicherungslauf");
					try {
						archiveManager.getInQueuesMgr().insertInBackupQueue(BackupTask.getBackupResultData());
						backupLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start des Sicherungslaufs", e);
					}
				}
				else if(tic == deleteTimeInterval) {
					logger.info("Starte reguläres Loeschen");
					try {
						archiveManager.getInQueuesMgr().insertInDeleteRegularQueue(LZVTask.getDeleteRegularResultData());
						deleteLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start des des regulären Löschens", e);
					}
				}
				else if(tic == deletePermanentlyTimeInterval) {
					logger.info("Starte endgueltiges Loeschen");
					try {
						archiveManager.getInQueuesMgr().insertInDeleteRegularQueue(LZVTask.getDeletePermanentlyResultData());
						deletePermanentlyLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start des des endgültigen Löschens", e);
					}
				}
				else if(tic == requestTimeInterval) {
					logger.info("Starte Nachforderung fehlender Archivdaten");
					try {
						archiveManager.getInQueuesMgr().insertInRequestQueue(RequestGapTask.getRequestGapResultData());
						requestLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start der Nachforderung fehlender Archivdaten", e);
					}
				}
			}
		}
	}
}
