/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import static de.bsvrz.ars.ars.persistence.ContainerFile.COMPRESS_LEN;
import static de.bsvrz.ars.ars.persistence.ContainerFile.DATAHDR_LEN;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NOT_COMPRESSED;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NO_DATA;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NO_RIGHTS;
import static de.bsvrz.ars.ars.persistence.ContainerFile.NO_SOURCE;
import static de.bsvrz.ars.ars.persistence.ContainerFile.POT_GAP;
import static de.bsvrz.ars.ars.persistence.ContainerFile.VERSION_STRING;


import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.util.CloseableRandomAccessFile;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import de.bsvrz.sys.funclib.losb.util.Util;
import java.io.File;
import java.io.IOException;
import java.util.LinkedHashSet;
import java.util.NoSuchElementException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.zip.Inflater;

/**
 * Diese Klasse liest aus einem einzelnen Container Datensätze
 *
 * @author Kappich Systemberatung
 */
public final class ContainerDataIterator implements DataIterator {

	/** true, wenn über das Ende hinaus gelesen wurde */
	private boolean _finished;

	/** Aktuelle Container-Datei */
	private final BasicContainerFileHandle _containerFileHandle;

	private final ContainerDataResult _next = new ContainerDataResult();

	private boolean _dataRead;

	private final Inflater _inflater;

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variablen für den Dateniterator:

	/** Puffer zum Einlesen von Längenbytes etc. */
	private final byte[] byte8Buf = new byte[8];

	/** Container-Datei. Ist ein RandomAccessFile, da Rückspruenge noetig sind, wenn der letzte Datensatz nur teilweise in den Puffer eingelesen wurde. */
	private final MyCloseableRandomAccessFile _randomAccessFile;

	/** Länge des nächsten Datensatzes (wird in iterReadNextData() besetzt) */
	private int _nextDataLen;
	
	private static final ThreadLocal<LinkedHashSet<MyCloseableRandomAccessFile>> _openFilesInCurrentThread = ThreadLocal.withInitial(LinkedHashSet::new);

	private static final AtomicInteger _numOpenFiles = new AtomicInteger(0);

	private int _counter;

	/**
	 * Maximalanzahl der gleichzeitig offenen Contaienrdateien
	 */
	private static int OPEN_FILE_LIMIT = 500;

	/**
	 * Gateilänge, wird hier aus Performancegründen gecacht, 
	 * da _randomAccessFile.length() laut profiling wesentlich Zeit benötigt. 
	 */
	private final long _length;

	public static int getOpenFileLimit() {
		return OPEN_FILE_LIMIT;
	}

	public static void setOpenFileLimit(final int openFileLimit) {
		OPEN_FILE_LIMIT = openFileLimit;
	}

	/**
	 * Erstellt einen neuen DataIteratorAdapter
	 * @param containerFileHandle Datenquelle
	 * @param inflater            Klasse zum Dekomprimieren (Optional)
	 */
	ContainerDataIterator(final BasicContainerFileHandle containerFileHandle, @Nullable final Inflater inflater) throws PersistenceException {
		_containerFileHandle = containerFileHandle;
		_inflater = inflater;
		_randomAccessFile = new MyCloseableRandomAccessFile(containerFileHandle.getFile());
		try {
			_length = _randomAccessFile.length();
			skipContainerHeader();
			_finished = !readNextDataHdr();
		}
		catch(IOException e) {
			throw new PersistenceException("Container " + containerFileHandle + " ist nicht lesbar", e);
		}
		if(_finished) {
			close();
		}
	}

	/**
	 * Springt im übergebenen Eingabestrom an die Stelle hinter dem Header der Container-Datei. Voraußetzung ist, dass das RandomAccessFile am Anfang der
	 * Datei steht.
	 *
	 * @throws IOException Allgemeiner IO-Fehler
	 */
	private void skipContainerHeader() throws IOException {
		_randomAccessFile.readFully(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
		_randomAccessFile.skipBytes(ByteIO.readSignedInt4Bytes(byte8Buf, VERSION_STRING.length));
	}

	@Override
	public void peek(final ContainerDataResult result) throws PersistenceException {
		if(_finished) {
			throw new NoSuchElementException();
		}
		if(_next.getData() == null && _inflater != null){
			try {
				_next.setData(readRawData());
			}
			catch(IOException e) {
				throw new PersistenceException("Datensatz im Container " + _containerFileHandle + " ist nicht lesbar", e);
			}
			_dataRead = true;
		}
		_next.copyTo(result);
	}

	@Nullable
	@Override
	public ContainerDataResult peekNext() {
		return null;
	}

	@Override
	public long peekDataIndex() {
		if(_finished) {
			throw new NoSuchElementException();
		}
		return _next.getDataIndex();
	}

	@Override
	public long peekDataTime() {
		if(_finished) {
			throw new NoSuchElementException();
		}
		return _next.getDataTime();
	}

	@Override
	public long peekArchiveTime() {
		if(_finished) {
			throw new NoSuchElementException();
		}
		return _next.getArchiveTime();
	}

	@Override
	public void remove() throws PersistenceException {
		if(_finished) {
			throw new NoSuchElementException();
		}
		if(!_dataRead) {
			try {
				skipData();
			}
			catch(IOException e) {
				throw new PersistenceException("Datensatz im Container " + _containerFileHandle + " konnte nicht übersprungen werden", e);
			}
		}
		else {
			_next.setData(null);
			_dataRead = false;
		}
		try {
			_finished = !readNextDataHdr();
		}
		catch(IOException e) {
			throw new PersistenceException("Nächster Datensatz im Container " + _containerFileHandle + " ist nicht lesbar", e);
		}
		if(_finished) {
			close();
		}
	}

	@Override
	public boolean isEmpty() {
		return _finished;
	}

	@Override
	public void close() {
		try {
			_randomAccessFile.close();
		}
		catch(IOException e) {
			Debug.getLogger().error("Container konnte nicht geschlossen werden: " + _containerFileHandle, e);
		}
		_next.setData(null);
	}

	@Override
	public BasicContainerFileHandle getContainerManagementData() throws PersistenceException {
		_containerFileHandle.ensureHeaderRead();
		return _containerFileHandle;
	}

	/**
	 * Liest den Datensatzheader
	 *
	 * @return true wenn erfolgreich gelesen, false wenn am Containerende
	 *
	 * @throws IOException          Wenn beim Lesen ein Fehler aufgetreten ist
	 * @throws PersistenceException Wenn beim Lesen ein Fehler aufgetreten ist
	 */
	private boolean readNextDataHdr() throws IOException, PersistenceException {
		if (_randomAccessFile.position() == _length) return false; // EOF

		_nextDataLen = _randomAccessFile.readInt();

		_next.setDataKind(_containerFileHandle.getLocation().getArchiveDataKind());
		_next.setContainerID(_containerFileHandle.getContainerId());

		_next.setArchiveTime(_randomAccessFile.read6ByteUnsigned());
		_next.setDataTime(_randomAccessFile.read6ByteUnsigned());
		_next.setDataIndex(_randomAccessFile.readLong());
		int checkSum = _randomAccessFile.readInt();

		int calcCheckSum = (int) (_nextDataLen + _next.getArchiveTime() + _next.getDataTime() + _next.getDataIndex());
		if(checkSum != calcCheckSum) {
			throw new IOException(
					"Prüfsumme des Datensatzes an Pos. " + _counter + " fehlerhaft: " + checkSum + " != dx=" + _next.getDataIndex()
							+ "+dt=" + _next.getDataTime() + "+at=" + _next.getArchiveTime() + "+len=" + _nextDataLen + "=" + calcCheckSum + "\n" + toString()
			);
		}

		_next.setDataSize(_nextDataLen - COMPRESS_LEN - DATAHDR_LEN - ByteIO.SEPARATOR.length);
		int dataUncompressedSize = _randomAccessFile.readInt();
		_next.setDataUncompressedSize(dataUncompressedSize);
		_next.setCompressed(dataUncompressedSize != NOT_COMPRESSED);
		_next.setStillCompressed(_next.isCompressed()); // das Archiv dekomprimiert nicht
		DataState state;
		state = DataState.DATA;
		if(_next.getDataSize() == 6) {
			long savedPosition = _randomAccessFile.position();
			byte[] buffer = new byte[6];
			_randomAccessFile.readFully(buffer);
			_randomAccessFile.position(savedPosition);
			if(Util.cmpBytes(buffer, 0, NO_DATA)) {
				state = DataState.NO_DATA;
			}
			else if(Util.cmpBytes(buffer, 0, NO_SOURCE)) {
				state = DataState.NO_SOURCE;
			}
			else if(Util.cmpBytes(buffer, 0, NO_RIGHTS)) {
				state = DataState.NO_RIGHTS;
			}
			else if(Util.cmpBytes(buffer, 0, POT_GAP)) {
				state = DataState.POSSIBLE_GAP;
			}
		}
		_next.setDataState(state);

		_counter++;
		return true;
	}


	@Nullable
	private byte[] readRawData() throws IOException {
		int netDataLen = _nextDataLen - COMPRESS_LEN - DATAHDR_LEN - ByteIO.SEPARATOR.length;

		if(!_next.isData()) {
			_randomAccessFile.skip(netDataLen + ByteIO.SEPARATOR.length);
			return null;
		}
		else {
			byte[] buf = new byte[netDataLen];
			_randomAccessFile.readFully(buf);
			_randomAccessFile.skip(ByteIO.SEPARATOR.length);
			return buf;
		}
	}

	private void skipData() throws IOException {
		int netDataLen = _nextDataLen - COMPRESS_LEN - DATAHDR_LEN;
		_randomAccessFile.skip(netDataLen);
	}


	private static class MyCloseableRandomAccessFile extends CloseableRandomAccessFile {

		MyCloseableRandomAccessFile(final File file) {
			super(file);
		}

		@Override
		protected void openFile() throws IOException {
			while(_numOpenFiles.get() > OPEN_FILE_LIMIT) {
				MyCloseableRandomAccessFile first = CollectionUtilities.removeOldest(_openFilesInCurrentThread.get());
				if(first != null) {
					// Braucht nicht aus Queue entfernt zu werden, da wir pollFirst() gemacht haben
					first.closeFile(false);
				}
				else {
					break;
				}
			}

			// Offene Dateien zählen
			_numOpenFiles.incrementAndGet();
			_openFilesInCurrentThread.get().add(this);
			super.openFile();
		}

		@Override
		public void closeFile() throws IOException {
			closeFile(true);
		}

		void closeFile(final boolean removeFromQueue) throws IOException {
			if(!isOpen()) return;
			// Datei nicht mehr zählen, da geschlossen
			_numOpenFiles.decrementAndGet();
			if(removeFromQueue) {
				_openFilesInCurrentThread.get().remove(this);
			}
			super.closeFile();
		}

	
	long read6ByteUnsigned() throws IOException {
	 return ((readUnsignedShort() & 0xFFFFL) << 32) | (readInt() & 0xFFFFFFFFL) ;
		}
	}
}
