/*
 * Archivsystem (ArS)
 * Copyright 2018-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import static de.bsvrz.ars.ars.persistence.ContainerFile.VERSION_STRING;


import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.dav.daf.util.BufferedRandomAccessFile;
import de.bsvrz.dav.daf.util.FileAccess;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import java.io.IOException;
import java.util.ArrayList;
import java.util.NoSuchElementException;

/**
 * Iterator über gelöschte Containerdateien, wird nur beim Upgrade einer bisherigen verkürzten Containerdatei verwendet.
 *
 * @author Kappich Systemberatung
 */
public class DeletedContainerDataIterator implements DataIterator {

	private final BasicContainerFileHandle _containerFileHandle;

	private int _counter;

	private final ArrayList<Long> _listOfIndexesDeleted;

	public DeletedContainerDataIterator(final BasicContainerFileHandle containerFileHandle) throws PersistenceException {
		_containerFileHandle = containerFileHandle;
		try (BufferedRandomAccessFile file = new BufferedRandomAccessFile(containerFileHandle.getFile())) {
			skipContainerHeader(file);
			_listOfIndexesDeleted = getListOfIndexesDeleted(file);
		}
		catch(IOException e) {
			throw new PersistenceException("Container " + containerFileHandle + " ist nicht lesbar", e);
		}
	}

	@Override
	public void peek(final ContainerDataResult result) {
		if(_counter >= _listOfIndexesDeleted.size()) throw new NoSuchElementException();
		result.setDataIndex(_listOfIndexesDeleted.get(_counter));
	}

	@Nullable
	@Override
	public ContainerDataResult peekNext() {
		return null;
	}

	@Override
	public long peekDataIndex() {
		if(_counter >= _listOfIndexesDeleted.size()) throw new NoSuchElementException();
		return _listOfIndexesDeleted.get(_counter);
	}

	@Override
	public long peekDataTime() {
		if(_counter >= _listOfIndexesDeleted.size()) throw new NoSuchElementException();
		return -1;
	}

	@Override
	public long peekArchiveTime() {
		if(_counter >= _listOfIndexesDeleted.size()) throw new NoSuchElementException();
		return -1;
	}

	@Override
	public void remove() {
		if(_counter >= _listOfIndexesDeleted.size()) throw new NoSuchElementException();
		_counter++;
	}

	@Override
	public boolean isEmpty() {
		return _counter >= _listOfIndexesDeleted.size();
	}

	@Override
	public void close() {
	}

	@NotNull
	@Override
	public ContainerManagementData getContainerManagementData() {
		return _containerFileHandle;
	}

	/**
	 * Springt im übergebenen Eingabestrom an die Stelle hinter dem Header der Container-Datei. Voraussetzung ist, dass das RandomAccessFile am Anfang der
	 * Datei steht.
	 *
	 * @param file RandomAccessFile der Container-Datei
	 * @throws IOException Allgemeiner IO-Fehler
	 */
	private static void skipContainerHeader(final BufferedRandomAccessFile file) throws IOException {
		byte[] byte8Buf = new byte[8];
		file.readFully(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
		file.skipBytes(ByteIO.readSignedInt4Bytes(byte8Buf, VERSION_STRING.length));
	}

	/**
	 * Erzeugt eine Liste der Datensatz-Indizes, die im durch raf spezifizierten Container vorkommen (sortiert)
	 *
	 * @param raf
	 * @throws IOException Allgemeiner IO-Fehler
	 */
	private static ArrayList<Long> getListOfIndexesDeleted(FileAccess raf) throws IOException {
		final ArrayList<Long> iterlistOfDataIdx = new ArrayList<>();
		byte[] first = new byte[ByteIO.LONG8B_LEN];
		byte[] last = new byte[ByteIO.LONG8B_LEN];

		while(raf.position() != raf.length()) {
			raf.readFully(first);
			if(raf.readByte() != '-') {
				// Separator
				throw new IOException("Erwarte Separator '-' an Dateiposition " + (raf.position()-1));
			} 
			raf.readFully(last);
			long firstIdx = ByteIO.readSignedLong8Byte(first, 0);
			long lastIdx = ByteIO.readSignedLong8Byte(last, 0);
			for(long i = firstIdx; i <= lastIdx; ) {
				iterlistOfDataIdx.add(i);
				i += 4;
			}
		}
		return iterlistOfDataIdx;
	}
}
