/*
 * Copyright 2019-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.gap.util;

import com.google.common.collect.Range;
import com.google.common.collect.RangeMap;
import com.google.common.collect.TreeRangeMap;
import com.google.common.collect.TreeRangeSet;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Eine {@link RangeMap} als {@link com.google.common.collect.Multimap}.
 *
 * @author Kappich Systemberatung
 */
public class RangeMultimap<K extends Comparable<? super K>, V extends Comparable<V>> {
	
	private final RangeMap<K, TreeSet<V>> _delegate = TreeRangeMap.create();

	@Nullable
	public SortedSet<V> get(final K key) {
		TreeSet<V> s = _delegate.get(key);
		if(s == null) return null;
		return Collections.unmodifiableSortedSet(s);
	}

	public Range<K> span() {
		return _delegate.span();
	}

	public void put(final Range<K> range, final V value) {
		addValueTo(value, range);
	}

	public void clear() {
		_delegate.clear();
	}

	public void remove(final Range<K> range) {
		_delegate.remove(range);
	}

	public void remove(final Range<K> range, final V value) {
		removeValueFrom(value, range);
	}

	private void addValueTo(final V value, final Range<K> span) {
		K lowerEndPoint = span.lowerEndpoint();
		K upperEndpoint = span.upperEndpoint();
		Map.Entry<Range<K>, TreeSet<V>> lowerEntry = _delegate.getEntry(lowerEndPoint);
		Map.Entry<Range<K>, TreeSet<V>> upperEntry = _delegate.getEntry(upperEndpoint);
		
		if(lowerEntry != null && !lowerEntry.getValue().contains(value) && !lowerEntry.getKey().lowerEndpoint().equals(lowerEndPoint)) {
			_delegate.put(lowerEntry.getKey().intersection(span), new TreeSet<>(lowerEntry.getValue()));
		}
		
		if(upperEntry != null && !upperEntry.getValue().contains(value) && !upperEntry.getKey().upperEndpoint().equals(upperEntry)) {
			_delegate.put(upperEntry.getKey().intersection(span), new TreeSet<>(upperEntry.getValue()));
		}

		TreeRangeSet<K> remainingSpans = TreeRangeSet.create();
		remainingSpans.add(span);
		for(Map.Entry<Range<K>, TreeSet<V>> entry : _delegate.subRangeMap(span).asMapOfRanges().entrySet()) {
			TreeSet<V> set = entry.getValue();
			set.add(value);
			remainingSpans.remove(entry.getKey());
		}
		for(Range<K> range : remainingSpans.asRanges()) {
			TreeSet<V> set = new TreeSet<>();
			set.add(value);
			_delegate.put(range, set);
		}
	}

	private void removeValueFrom(final V value, final Range<K> span) {
		K lowerEndPoint = span.lowerEndpoint();
		K upperEndpoint = span.upperEndpoint();
		Map.Entry<Range<K>, TreeSet<V>> lowerEntry = _delegate.getEntry(lowerEndPoint);
		Map.Entry<Range<K>, TreeSet<V>> upperEntry = _delegate.getEntry(upperEndpoint);

		if(lowerEntry != null && lowerEntry.getValue().contains(value) && !lowerEntry.getKey().lowerEndpoint().equals(lowerEndPoint)) {
			_delegate.put(lowerEntry.getKey().intersection(span), new TreeSet<>(lowerEntry.getValue()));
		}

		if(upperEntry != null && upperEntry.getValue().contains(value) && !upperEntry.getKey().upperEndpoint().equals(upperEntry)) {
			_delegate.put(upperEntry.getKey().intersection(span), new TreeSet<>(upperEntry.getValue()));
		}

		for(Iterator<TreeSet<V>> iterator = _delegate.subRangeMap(span).asMapOfRanges().values().iterator(); iterator.hasNext(); ) {
			final TreeSet<V> set = iterator.next();
			set.remove(value);
			if(set.isEmpty()) {
				iterator.remove();
			}
		}
	}

	public Map<Range<K>, TreeSet<V>> asMapOfRanges() {
		return _delegate.asMapOfRanges();
	}

	public Map<Range<K>, TreeSet<V>> asDescendingMapOfRanges() {
		return _delegate.asDescendingMapOfRanges();
	}

	public RangeMap<K, TreeSet<V>> subRangeMap(final Range<K> range) {
		return _delegate.subRangeMap(range);
	}

	@Override
	public String toString() {
		return _delegate.toString();
	}
}
