/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.ContainerManagementData;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.zip.Inflater;

/**
 * Iteriert wie der {@link SequentialContainerIterator} über eine Menge von Containern. Diese Klasse muss benutzt werden, wenn sich die Containerdateien
 * zeitlich bzw. datenindexmäßig überschneiden.
 * <p>
 * In diesem Fall, werden die Container-Dateien in Buckets sortiert, wobei ein Bucket immer eine Menge an sich überschneidenden Containerdateien enthält. Die
 * Buckets selbst werden aufsteigend nach Datenindex sortiert.
 * <p>
 * Diese Klasse erzeugt die Buckets nicht selbst, sondern sie werden im Konstruktor übergeben. Diese Klasse übernimmt dass iterieren und sortieren von den
 * einzelnen Datensätzen.
 *
 * @author Kappich Systemberatung
 */
public class BucketContainerIterator implements DataIterator {

	private final ArchiveManager _archiveManager;
	private final List<List<QueryDataSequence.IndexedContainer>> _containerFiles;
	private final SyncKey<IdDataIdentification> _indexLock;
	private final ArchiveDataKind _adk;
	private DataIterator _currentDelegate;
	private int _index;
	private final ArchiveOrder _order;
	private final DeletedTreatment _deletedTreatment;
	private final Inflater _inflater;

	/**
	 * Erstellt einen neuen BucketContainerIterator
	 * @param archiveManager   Archivmanager
	 * @param containerFiles   Vorsortierte Container-Dateien. Eine Liste von Container-Dateien-Gruppen, die sich jeweils zeitlich überschneiden. Die äußere Liste
	 *                         ist nach Datenindex sortiert.
	 * @param indexLock Datenidentifikation
	 * @param adk              Archivdatenart
	 * @param deletedTreatment Behandlung von gelöschten Daten
	 * @param order            Sortierung
	 * @param inflater         Komprimierung
	 */
	public BucketContainerIterator(final ArchiveManager archiveManager, final List<List<QueryDataSequence.IndexedContainer>> containerFiles, final SyncKey<IdDataIdentification> indexLock, final ArchiveDataKind adk, final DeletedTreatment deletedTreatment, final ArchiveOrder order, final Inflater inflater) throws PersistenceException, SynchronizationFailedException {
		_archiveManager = archiveManager;
		_containerFiles = containerFiles;
		_indexLock = indexLock;
		_adk = adk;
		_deletedTreatment = deletedTreatment;
		_order = order;
		_inflater = inflater;
		if(containerFiles.isEmpty()) throw new IllegalArgumentException("containerFiles");
		openBucket(_index);
	}


	private void openBucket(final int index) throws PersistenceException, SynchronizationFailedException {
		List<QueryDataSequence.IndexedContainer> bucket = _containerFiles.get(index);
		if(bucket.size() == 1) {
			_currentDelegate = new SequentialContainerIterator(_archiveManager, bucket, _adk, _deletedTreatment, null, _inflater,_indexLock);
		}
		else {
			final List<DataIterator> iterators = new ArrayList<>();
			for(QueryDataSequence.IndexedContainer indexedContainer : bucket) {
				iterators.add(new SequentialContainerIterator(_archiveManager, Collections.singletonList(indexedContainer), _adk, _deletedTreatment, null, _inflater, _indexLock));
			}
			_currentDelegate = new CombineDataIterator(iterators, _order);
		}
	}

	private void closeBucket() {
		_currentDelegate.close();
	}

	@Override
	public void peek(final ContainerDataResult result) throws PersistenceException {
		_currentDelegate.peek(result);
	}

	@Override
	public ContainerDataResult peekNext() {
		if(!isEmpty()) {
			throw new IllegalStateException();
		}
		
		return null;
	}

	@Override
	public long peekDataIndex() {
		return _currentDelegate.peekDataIndex();
	}

	@Override
	public long peekDataTime() {
		return _currentDelegate.peekDataTime();
	}

	@Override
	public long peekArchiveTime() {
		return _currentDelegate.peekArchiveTime();
	}

	@Override
	public void remove() throws PersistenceException, SynchronizationFailedException {
		_currentDelegate.remove();
		if(_currentDelegate.isEmpty() && _index < _containerFiles.size() - 1) {
			closeBucket();
			_index++;
			openBucket(_index);
		}
	}

	@Override
	public boolean isEmpty() {
		return _currentDelegate.isEmpty();
	}

	@Override
	public void close() {
		_currentDelegate.close();
	}

	@Override
	public ContainerManagementData getContainerManagementData() throws PersistenceException {
		if(_currentDelegate == null) throw new NoSuchElementException();
		return _currentDelegate.getContainerManagementData();
	}
}
