/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SyncKey;
import de.bsvrz.ars.ars.mgmt.datatree.synchronization.SynchronizationFailedException;
import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.ContainerManagementData;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import java.util.List;
import java.util.zip.Inflater;

/**
 * Iteration über die Daten von einer Liste von Containern. Die Container müssen bereits nach Datenindex nicht überlappend aufsteigend sortiert sein.
 *
 * Ist das nicht der Fall (z.B. bei nachgeforderten Daten), sollte ein {@link BucketContainerIterator} benutzt werden.
 *
 * @author Kappich Systemberatung
 */
public class SequentialContainerIterator implements DataIterator {

	private final List<QueryDataSequence.IndexedContainer> _containerFiles;
	private final ArchiveManager _archiveManager;
	private DataIterator _dataIterator;
	private final DeletedTreatment _deletedTreatment;
	private int _index;
	private final ArchiveDataKind _adk;
	private final QueryDataSequence.IndexedContainer _nextContainer;
	private final Inflater _inflater;
	private final SyncKey<IdDataIdentification> _indexLock;

	/**
	 * Erstellt einen neuen SequentialContainerIterator
	 * @param archiveManager Archivmanager
	 * @param containerFiles Containerdateien, bereits sortiert
	 * @param adk              Archivdatenart
	 * @param deletedTreatment Behandlung von gelöschten Containern
	 * @param nextContainer    Den Containerdateien folgender Container
	 * @param inflater         Entpacker für Daten (Unzip)
	 * @param indexLock        Lock für Zugriff auf Indexe
	 * @throws PersistenceException Lesefehler im Persistenzverzeichnis
	 */
	public SequentialContainerIterator(final ArchiveManager archiveManager, final List<QueryDataSequence.IndexedContainer> containerFiles, final ArchiveDataKind adk, final DeletedTreatment deletedTreatment, @Nullable final QueryDataSequence.IndexedContainer nextContainer, final Inflater inflater, final SyncKey<IdDataIdentification> indexLock) throws PersistenceException, SynchronizationFailedException {
		_archiveManager = archiveManager;
		_adk = adk;
		_deletedTreatment = deletedTreatment;
		_nextContainer = nextContainer;
		_inflater = inflater;
		_indexLock = indexLock;
		if(containerFiles.isEmpty()) throw new IllegalArgumentException("containerFiles");
		_containerFiles = containerFiles;
		openContainer(_index);
	}

	private void openContainer(final int index) throws PersistenceException, SynchronizationFailedException {
		QueryDataSequence.IndexedContainer container = _containerFiles.get(index);
		_dataIterator = getIterator(container);
	}

	@Override
	public void peek(final ContainerDataResult result) throws PersistenceException {
		_dataIterator.peek(result);
	}

	@Override
	public ContainerDataResult peekNext() throws PersistenceException, SynchronizationFailedException {
		if(!isEmpty()) {
			throw new IllegalStateException();
		}
		ContainerDataResult peekNext = _dataIterator.peekNext();
		if(peekNext == null && _nextContainer != null) {
			try(DataIterator iterator = getIterator(_nextContainer)) {
				ContainerDataResult result = new ContainerDataResult();
				iterator.poll(result);
				return result;
			}
		}
		return peekNext;
	}

	private DataIterator getIterator(final QueryDataSequence.IndexedContainer nextContainer) throws PersistenceException, SynchronizationFailedException {
		return _archiveManager.getPersMgr().iterator(_adk, nextContainer.getContainerId(), _deletedTreatment, _inflater, _indexLock);
	}

	@Override
	public long peekDataIndex() {
		return _dataIterator.peekDataIndex();
	}

	@Override
	public long peekDataTime() {
		return _dataIterator.peekDataTime();
	}

	@Override
	public long peekArchiveTime() {
		return _dataIterator.peekArchiveTime();
	}

	@Override
	public void remove() throws PersistenceException, SynchronizationFailedException {
		_dataIterator.remove();
		while(_dataIterator.isEmpty() && _index < _containerFiles.size() - 1) {
			_dataIterator.close();
			_index++;
			openContainer(_index);
		}
	}

	@Override
	public boolean isEmpty() {
		return _dataIterator.isEmpty();
	}

	@Override
	public void close() {
		_dataIterator.close();
	}

	@Override
	public ContainerManagementData getContainerManagementData() throws PersistenceException {
		return _dataIterator.getContainerManagementData();
	}
}
