/*
 *
 * Copyright 2017-2020 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.ars.ars.
 *
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.walk.internal;

import java.text.NumberFormat;
import java.time.Duration;


/**
 * Klasse, die den Status einer ContainerDirWalk-Aktion ausgibt. Kann überschrieben werden um einen angepassten/erweiterten Status zu veröffentlichen.
 *
 * @author Kappich Systemberatung
 */
public class StatusPrinter {

	/**
	 * Gibt die Auszugebende Statusmeldung zurück. Kann überschrieben werden um weitere Informationen hinzuzufügen.
	 * @param actionName Name der Aktion / des Tasks
	 * @param runtime Bisherige Laufzeit
	 * @param indexFinished Sind bereits alle relevanten Dateien indiziert?
	 * @param indexedContainerDirs Anzahl der bereits indizierten Containerdateien-Ordner
	 * @param visitedContainerDirs Anzahl der bereits besuchten Containerdateien-Ordner
	 * @param visitedContainerFiles
	 * @return Nachricht
	 */
	public String getStatusMessage(final String actionName, final Duration runtime, final boolean indexFinished, final long indexedContainerDirs, final long visitedContainerDirs, final long visitedContainerFiles) {
		if(!indexFinished) {
			return actionName + " läuft.\nBisherige Laufzeit: " + formatDuration(runtime) + "...\n"
					+ visitedContainerDirs + " von mindestens " + indexedContainerDirs + " Container-Verzeichnissen besucht.\n"
					+ estimateMinRuntime(runtime, visitedContainerDirs, indexedContainerDirs) + "..."
					+ (visitedContainerFiles == 0 ? "" :"\n" + visitedContainerFiles + " Container-Dateien besucht.");
		}
		else {
			return actionName + " läuft.\nBisherige Laufzeit: " + formatDuration(runtime) + "...\n"
					+ visitedContainerDirs + " von " + indexedContainerDirs + " Container-Verzeichnissen besucht.\n"
					+ estimateRuntime(runtime, visitedContainerDirs, indexedContainerDirs) + "..."
					+ (visitedContainerFiles == 0 ? "" :"\n" + visitedContainerFiles + " Container-Dateien besucht.");

		}
	}

	/**
	 * Formatiert die Ausgabe für die Fertigmeldung.
	 * @param actionName           Name der Aktion
	 * @param runtime Bisherige Laufzeit
	 * @param visitedContainerDirs Anzahl besuchter Containerdateien-Verzeichnisse
	 * @param visitedContainerFiles Anzahl besuchte Containerdateien
	 * @return Textuelle Ausgabe
	 */
	public String getSuccessMessage(final String actionName, final Duration runtime, final long visitedContainerDirs, final long visitedContainerFiles) {
		return actionName + " beendet.\nLaufzeit: " + formatDuration(runtime) + ".\n" + visitedContainerDirs + " Container-Verzeichnisse besucht." +
		       (visitedContainerFiles == 0 ? "" : "\n" + visitedContainerFiles + " Container-Dateien besucht.");

	}

	/**
	 * Berechnet den Fortschritt und formatiert eine Ausgabe.
	 * @param runtime Bisherige Laufzeit
	 * @param visited Anzahl besuchter Objekte
	 * @param all Anzahl aller Objekte
	 * @return Textuelle Ausgabe
	 */
	public static String estimateRuntime(final Duration runtime, final long visited, final long all) {
		double finishedRatio = visited / (double)all;
		long currentRuntimeSeconds = runtime.getSeconds();
		double estimatedCompleteTime = currentRuntimeSeconds / finishedRatio;
		if(all == 0 || estimatedCompleteTime > Duration.ofDays(1000).getSeconds()) {
			// Überlauf vermeiden
			return "Fortschritt: Unbekannt";
		}
		long remainingSeconds = (long) (estimatedCompleteTime - currentRuntimeSeconds);
		Duration remainingDuration = Duration.ofSeconds(remainingSeconds);
		String duration = formatDuration(remainingDuration);
		return "Fortschritt: " + NumberFormat.getPercentInstance().format(finishedRatio) + ", noch ca. " + duration;
	}

	/**
	 * Berechnet den Fortschritt und formatiert eine Ausgabe wenn die Gesamtzahl der Objekte unbekannt ist.
	 * @param runtime Bisherige Laufzeit
	 * @param visited Anzahl besuchter Objekte
	 * @param indexed Anzahl bisher gefundener Objekte
	 * @return Textuelle Ausgabe
	 */
	public static String estimateMinRuntime(final Duration runtime, final long visited, final long indexed) {
		double maxFinishedRatio = visited / (double)indexed;
		long currentRuntimeSeconds = runtime.getSeconds();
		double estimatedMinCompleteTime = currentRuntimeSeconds / maxFinishedRatio;
		long remainingMinSeconds = (long) (estimatedMinCompleteTime - currentRuntimeSeconds);
		Duration remainingDuration = Duration.ofSeconds(remainingMinSeconds);
		if(remainingDuration.toMinutes() < 1) {
			return "Fortschritt: Unbekannt";
		}
		String minDuration = formatDuration(remainingDuration);
		return "Fortschritt: Unbekannt, geschätzte Laufzeit: Noch mindestens " + minDuration;
	}

	/**
	 * Formatiert eine Zeitdauer.
	 * @param runtime Dauer
	 * @return Formatierte Ausgabe
	 */
	private static String formatDuration(final Duration runtime) {
		long days = runtime.toDays();
		long hours = runtime.toHours() % 24;
		long minutes = runtime.toMinutes() % 60;
		long seconds = runtime.getSeconds() % 60;
		if(days == 0) return String.format("%02d:%02d:%02d", hours, minutes, seconds);
		if(days == 1) return String.format("1 Tag, %02d:%02d:%02d", hours, minutes, seconds);

		return String.format("%d Tage, %02d:%02d:%02d", days, hours, minutes, seconds);
	}
}
